# from jaqpotpy.docking.pose_generation import PoseGenerator
# from jaqpotpy.docking.pose_generation import VinaPoseGenerator
# from jaqpotpy.docking.pose_generation import GninaPoseGenerator
# from jaqpotpy.docking.docking import Docker
# from jaqpotpy.docking.binding_pocket import ConvexHullPocketFinder



"""Author: Ioannis Pitoskas (jpitoskas@gmail.com)"""

import torch.nn as nn
from typing import Iterable, Union
import torch.nn.init as init
from torch import Tensor


class FullyConnectedBlock(nn.Module):
    """A single fully connected block consisting of a linear layer, an activation function,
    and a dropout layer. Xavier initialization is applied to the weights.

    Attributes
    ----------
        fc (nn.Linear): The Fully Connected linear layer.
        activation (nn.Module): The activation function.
        dropout (nn.Dropout): The dropout layer.

    """

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        activation: nn.Module = nn.ReLU(),
        dropout_probability: float = 0.5,
        #  norm: Optional[bool] = False,
        *args,
        **kwargs,
    ):
        """Args:
        ----
            input_dim (int): Dimension of the input features.
            output_dim (int): Dimension of the output features.
            activation (nn.Module): Activation function to apply after the linear layer.
            dropout_probability (float): Dropout probability.

        """
        super().__init__()

        self.fc = nn.Linear(input_dim, output_dim)
        self.activation = activation
        self.dropout = nn.Dropout(p=dropout_probability)

        # Apply Xavier initialization to fc
        init.xavier_uniform_(self.fc.weight)
        init.zeros_(self.fc.bias)

    def forward(self, x: Tensor) -> Tensor:
        """Passes the input through the layer.

        Args:
        ----
            x (Tensor): Input tensor.

        Returns:
        -------
            Tensor: Output tensor after applying the fully connected block.

        """
        x = self.fc(x)
        x = self.activation(x)
        x = self.dropout(x)

        return x


class FullyConnectedNetwork(nn.Module):
    """A fully connected neural network consisting of multiple FullyConnectedBlock layers
    and a final linear projection. The network can be customized with various activation
    functions and dropout probabilities for each layer.

    Attributes
    ----------
        input_dim (int): Dimension of the input features.
        output_dim (int): Dimension of the output features.
        hidden_dims (Iterable[int]): Dimensions of the hidden layers.
        activation (nn.Module): Activation function to apply after each hidden layer.
        dropout_probabilities (list): List of dropout probabilities after each FC layer
        fc_layers (nn.ModuleList): ModuleList of FullyConnectedBlock layers of the network
        fc_head (nn.Linear): The final linear projection layer.

    """

    def __init__(
        self,
        input_dim: int,
        hidden_dims: Iterable[int],
        output_dim: int = 1,
        activation: nn.Module = nn.ReLU(),
        dropout: Union[float, Iterable[float]] = 0.5,
        #  norm: Optional[bool] = False,
        *args,
        **kwargs,
    ):
        """Args:
        ----
            input_dim (int): Dimension of the input features.
            hidden_dims (Iterable[int]): Dimensions of the hidden layers.
            output_dim (int): Dimension of the output features. Default is 1.
            activation (nn.Module): Activation function to apply after each hidden layer. Default is nn.ReLU().
            dropout (Union[float, Iterable[float]]): Dropout probabilities for each hidden layer. Default is 0.5.

        """
        super().__init__()

        # Input types check
        if not isinstance(input_dim, int):
            raise TypeError("input_dim must be of type int")

        if not isinstance(hidden_dims, Iterable):
            raise TypeError("hidden_dims must be an Iterable")

        if not all(isinstance(hidden_dim, int) for hidden_dim in hidden_dims):
            raise TypeError("hidden_dims must only contain integers")

        if not isinstance(output_dim, int):
            raise TypeError("output_dim must be of type int")

        if not isinstance(activation, nn.Module):
            raise TypeError("activation must be a torch.nn.Module")

        if not (isinstance(dropout, float) or isinstance(dropout, Iterable)):
            raise TypeError("dropout must be either of type float or Iterable")
        if isinstance(dropout, float):
            if not 0 <= dropout <= 1:
                raise ValueError("dropout probability must be between 0 and 1")
        if isinstance(dropout, Iterable):
            for item in dropout:
                if not isinstance(item, float):
                    raise TypeError("dropout list must only contain floats")
                if not 0 <= item <= 1:
                    raise ValueError(
                        "Each element in the dropout list must be between 0 and 1"
                    )
            if len(dropout) != len(hidden_dims):
                raise ValueError("hidden_dims and dropout must be of same size")

        self.input_dim = input_dim
        self.output_dim = output_dim
        self.hidden_dims = hidden_dims
        self.dropout_probabilities = (
            [dropout] * len(hidden_dims) if isinstance(dropout, float) else dropout
        )

        if self.hidden_dims == []:
            self.fc_layers = None
            self.fc_head = nn.Linear(input_dim, output_dim)
        else:
            self.fc_layers = nn.ModuleList()
            fc_layer = FullyConnectedBlock(
                input_dim,
                self.hidden_dims[0],
                activation=activation,
                dropout_probability=self.dropout_probabilities[0],
            )
            self.fc_layers.append(fc_layer)

            for i in range(len(hidden_dims) - 1):
                fc_layer = FullyConnectedBlock(
                    self.hidden_dims[i],
                    hidden_dims[i + 1],
                    activation=activation,
                    dropout_probability=self.dropout_probabilities[i],
                )
                self.fc_layers.append(fc_layer)

            self.fc_head = nn.Linear(hidden_dims[-1], output_dim)

    def forward(self, x: Tensor) -> Tensor:
        """Passes the input through the network layers.

        Args:
        ----
            x (Tensor): Input tensor.

        Returns:
        -------
            Tensor: Output tensor after passing through the network.

        """
        if self.fc_layers is not None:
            for fc_layer in self.fc_layers:
                x = fc_layer(x)

        x = self.fc_head(x)

        return x



from jaqpotpy.descriptors.material.geometrical import GeomDescriptors
from jaqpotpy.descriptors.material.element_property_fingerprint import (
    ElementPropertyFingerprint,
)

from jaqpotpy.descriptors.material.cgcnn import CrystalGraphCNN
from jaqpotpy.descriptors.material.elemnet import ElementNet
from jaqpotpy.descriptors.material.sine_coulomb_matrix import SineCoulombMatrix

__all__ = ["GeomDescriptors", "ElementPropertyFingerprint",
           "CrystalGraphCNN", "ElementNet", "SineCoulombMatrix"]



import requests


doa_path = "doa"


def post_models_doa(baseurl, api_key, json_request, logger):
    uri = baseurl + doa_path
    h = {
        "Content-Type": "application/json",
        "Accept": "application/json",
        "Authorization": "Bearer " + api_key,
    }
    try:
        r = requests.post(uri, headers=h, data=json_request)
        return r.status_code
    except Exception as e:
        logger.error("Error http: " + str(e))


def get_models_doa(baseurl, api_key, modelId, logger):
    uri = baseurl + doa_path
    h = {
        "Content-Type": "application/json",
        "Accept": "application/json",
        "Authorization": "Bearer " + api_key,
    }
    d = {"hasSources": modelId}
    try:
        r = requests.post(uri, headers=h, data=d)
        return r.status_code
    except Exception as e:
        logger.error("Error http: " + str(e))



import json
import pandas as pd


def decode_predicted(dataset):
    feat_info = dataset["features"]
    predicted = {}
    predicted_f = []
    for feat in feat_info:
        name = feat["name"]
        key = feat["key"]
        try:
            category = feat["category"]
            if (category == "PREDICTED") and (feat["name"] not in predicted_f):
                predicted_f.append(feat["name"])
        except KeyError:
            dataentries = {}
            for dataEntry in dataset["dataEntry"]:
                dataentries[dataEntry["entryId"]["name"]] = dataEntry["values"][key]
            predicted[name] = dataentries
        dataentries = {}
        for dataEntry in dataset["dataEntry"]:
            dataentries[dataEntry["entryId"]["name"]] = dataEntry["values"][key]
        predicted[name] = dataentries
    json_to_df = json.dumps(predicted)
    df = pd.read_json(json_to_df)
    return df, predicted_f

    # for val in dataEntry['values']:


#         for dataEntry in dataset['dataEntry']:
#             dataEntryToInsert = []
#             for key in shorted:
#                 dataEntryToInsert.append(dataEntry['values'][key])
#             dataEntryAll.append(dataEntryToInsert)
#     for key in independentFeatures:
#         if actual == independentFeatures[key]:
#             for feature in pred_request.dataset['features']:
#                 if feature['name'] == actual:
#                     shorted.append(feature['key'])
#             # shorted.append(key)
# dataEntryAll = []
# for dataEntry in pred_request.dataset['dataEntry']:
#     dataEntryToInsert = []
#     for key in shorted:
#         dataEntryToInsert.append(dataEntry['values'][key])
#     dataEntryAll.append(dataEntryToInsert)
# return dataEntryAll



import requests

task_path = "task"


def get_task(baseurl, api_key, taskid):
    uri = baseurl + task_path + "/" + taskid
    h = {
        "Content-Type": "application/x-www-form-urlencoded",
        "Accept": "application/json",
        "Authorization": "Bearer " + api_key,
    }
    r = requests.get(uri, headers=h)
    return r.json()



"""Author: Ioannis Pitoskas (jpitoskas@gmail.com)"""


class Organization:
    def __init__(self):
        pass

    def to_json(self):
        organization_dict = {}
        return organization_dict

    def __str__(self):
        return self.to_json()

    def __repr__(self):
        return f"{self.__class__.__name__}({self.to_json()})"



from jaqpotpy.cfg import config
from jaqpotpy.datasets import JaqpotpyDataset
from jaqpotpy.descriptors.molecular import MolGanFeaturizer, GraphMatrix
import torch
from torch_geometric.nn import GCNConv
import torch.nn.functional as F
from torch_geometric.nn import global_mean_pool


import pandas as pd

from tensorflow import keras

from rdkit import RDLogger

RDLogger.DisableLog("rdApp.*")

csv_path = keras.utils.get_file(
    "./250k_rndm_zinc_drugs_clean_3.csv",
    "https://raw.githubusercontent.com/aspuru-guzik-group/chemical_vae/master/models/zinc_properties/250k_rndm_zinc_drugs_clean_3.csv",
)

df = pd.read_csv(
    "/Users/pantelispanka/.keras/datasets/250k_rndm_zinc_drugs_clean_3.csv"
)
df["smiles"] = df["smiles"].apply(lambda s: s.replace("\n", ""))
print(df.head())

smiles = df["smiles"].to_list()[0:1000]

feat = MolGanFeaturizer(max_atom_count=40)

dataset = JaqpotpyDataset(smiles=smiles, featurizer=feat, task="generation")
dataset = dataset.create()
gm: GraphMatrix = dataset.df["MolGanGraphs"][0]


class Generator(torch.nn.Module):
    def __init__(
        self,
        in_channels,
        num_layers,
        hidden_channels,
        out_channels,
        activation: torch.nn.Module = torch.nn.ReLU(),
        dropout=0,
        norm: torch.nn.Module = None,
        act_first=True,
    ):
        super(Generator, self).__init__()
        torch.manual_seed(config.global_seed)

        self.act = activation
        self.dropout = dropout
        self.act_first = act_first
        self.norm = norm

        self.layers = torch.nn.ModuleList([GCNConv(in_channels, hidden_channels)])
        for i in range(num_layers - 1):
            self.layers.append(GCNConv(hidden_channels, hidden_channels))
        self.out = torch.nn.Linear(hidden_channels, out_channels)

    def forward(self, data):
        x, edge_index = data.x, data.edge_index
        for layer in self.layers:
            x = layer(x, edge_index)
            if self.norm and self.act_first is True:
                x = self.act(x)
                x = self.norm(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
            elif self.norm and self.act_first is False:
                x = self.norm(x)
                x = self.act(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
            else:
                x = self.act(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
        x = global_mean_pool(x, data.batch)
        x = self.out(x)
        return x


class Discriminator(torch.nn.Module):
    def __init__(
        self,
        in_channels,
        num_layers,
        hidden_channels,
        out_channels,
        activation: torch.nn.Module = torch.nn.ReLU(),
        dropout=0,
        norm: torch.nn.Module = None,
        act_first=True,
    ):
        super(Discriminator, self).__init__()
        torch.manual_seed(config.global_seed)
        self.act = activation
        self.dropout = dropout
        self.act_first = act_first
        self.norm = norm
        self.layers = torch.nn.ModuleList([GCNConv(in_channels, hidden_channels)])
        for i in range(num_layers - 1):
            self.layers.append(GCNConv(hidden_channels, hidden_channels))
        self.out = torch.nn.Linear(hidden_channels, out_channels)

    def forward(self, data):
        x, edge_index = data.x, data.edge_index
        for layer in self.layers:
            x = layer(x, edge_index)
            if self.norm and self.act_first is True:
                x = self.act(x)
                x = self.norm(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
            elif self.norm and self.act_first is False:
                x = self.norm(x)
                x = self.act(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
            else:
                x = self.act(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
        x = global_mean_pool(x, data.batch)
        x = self.out(x)
        return x



"""Author: Ioannis Pitoskas (jpitoskas@gmail.com)"""

from abc import ABC, abstractmethod, ABCMeta
import torch
import logging
import sys

import requests
import os
from typing import List, Optional, Union
from jaqpotpy.schemas import Feature, Library, Organization
from jaqpotpy.utils.installed_packages import get_installed_packages
import inspect
from torch.optim.lr_scheduler import LambdaLR


class TorchModelTrainerMeta(ABCMeta):
    def __new__(cls, name, bases, dct):
        """A metaclass for torch model training, ensuring:
        - 'get_model_type' method is defined as a class method
        """
        if "get_model_type" in dct:
            method = dct["get_model_type"]
            if not isinstance(method, classmethod):
                raise TypeError(f"{name}.get_model_type must be a class method")
        return super().__new__(cls, name, bases, dct)


class TorchModelTrainer(ABC, metaclass=TorchModelTrainerMeta):
    """An abstract class for training a model and deploying it on Jaqpot.

    Attributes
    ----------
        model (torch.nn.Module): The torch model to be trained.
        n_epochs (int): Number of training epochs.
        optimizer (torch.optim.Optimizer): The optimizer used for training the model.
        loss_fn (torch.nn.Module): The loss function used for training.
        scheduler (torch.optim.lr_scheduler.LRScheduler): The scheduler used for adjusting the learning rate during training.
        device (torch.device): The device on which to train the model.
        use_tqdm (bool): Whether to use tqdm for progress bars.
        current_epoch (int): The epoch on which the trainer has currently reached.
        log_enabled (bool): Whether logging is enabled.
        log_filepath (os.path.relpath or None): Relative path to the log file.
        json_data_for_deployment (dict or None): The data to be sent to the API of Jaqpot in JSON format. Note that `prepare_for_deployment` must be called to compute this attribute.
        logger (logging.Logger): The logger object at INFO level used for logging during model training.

    """

    @classmethod
    @abstractmethod
    def get_model_type(cls):
        """Return the type of the model as a string.

        Returns
        -------
            str: The model type.

        """
        pass

    def __init__(
        self,
        model: torch.nn.Module,
        n_epochs: int,
        optimizer: torch.optim.Optimizer,
        loss_fn: torch.nn.Module,
        scheduler: Optional[torch.optim.lr_scheduler.LRScheduler] = None,
        device: str = "cpu",
        use_tqdm: bool = True,
        log_enabled: bool = True,
        log_filepath: Optional[str] = None,
    ):
        """Args:
        ----
            model (torch.nn.Module): The torch model to be trained.
            n_epochs (int): Number of training epochs.
            optimizer (torch.optim.Optimizer): The optimizer used for training the model.
            loss_fn (torch.nn.Module): The loss function used for training.
            scheduler (torch.optim.lr_scheduler.LRScheduler): The scheduler used for adjusting the learning rate during training. Default is None.
            device (str, optional): The device on which to train the model. Default is 'cpu'.
            use_tqdm (bool, optional): Whether to use tqdm for progress bars. Default is True.
            log_enabled (bool, optional): Whether logging is enabled. Default is True.
            log_filepath (str or None, optional): Path to the log file. If None, logging is not saved to a file. Default is None.

        """
        self.model = model
        self.n_epochs = n_epochs
        self.optimizer = optimizer
        self.scheduler = (
            scheduler
            if scheduler is not None
            else LambdaLR(optimizer, lr_lambda=lambda epoch: 1)
        )
        self.loss_fn = loss_fn
        self.device = torch.device(device)
        self.use_tqdm = use_tqdm
        self.current_epoch = 0
        self.log_enabled = log_enabled
        self.log_filepath = (
            os.path.relpath(log_filepath) if log_filepath else log_filepath
        )
        self.json_data_for_deployment = None

        self.logger = self._setup_logger()

        self.model.to(self.device)

    def _setup_logger(self):
        """Setup the logger for the training process."""
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.INFO)

        # Remove existing handlers
        for handler in logger.handlers[:]:
            logger.removeHandler(handler)

        handlers = [logging.StreamHandler(sys.stdout)]
        if self.log_filepath:
            if os.path.exists(self.log_filepath):
                raise ValueError(f"File already exists: {self.log_filepath}")
            log_file_handler = logging.FileHandler(self.log_filepath)
            handlers.append(log_file_handler)

        log_format = "%(message)s"
        formatter = logging.Formatter(log_format)
        for handler in handlers:
            handler.setFormatter(formatter)
            logger.addHandler(handler)

        return logger

    @classmethod
    def collect_subclass_info(cls):
        """Collect information about all subclasses of the current class.

        Returns
        -------
            dict: A dictionary containing information about the subclasses (model type, parent class, if is abstract class or not).

        """
        subclass_info = {}

        for subclass in cls.__subclasses__():
            subclass_info[subclass.__name__] = {
                "model_type": subclass.get_model_type(),
                "parent": cls.__name__,
                "is_abstract": inspect.isabstract(subclass),
            }

            subclass_info.update(subclass.collect_subclass_info())

        return subclass_info

    @classmethod
    def get_subclass_model_types(cls):
        """Return the list of all the types of models that inherit from TorchModelTrainer and are implemented in the jaqpotpy_torch.trainers submodule.

        Returns
        -------
            list: A list of model types as strings.

        """
        return [
            v["model_type"]
            for v in cls.collect_subclass_info().values()
            if not v["is_abstract"] and v["model_type"] is not None
        ]

    # @check_reach_epoch_limit
    @abstractmethod
    def train(self, train_loader, val_loader=None):
        """Train the model.

        Args:
        ----
            train_loader (Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]): DataLoader for the training dataset.
            val_loader (Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader], optional): DataLoader for the validation dataset.

        Returns:
        -------
            None

        """
        pass

    @abstractmethod
    def evaluate(self, val_loader):
        """Evaluate the model.

        Args:
        ----
            loader (Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]): DataLoader for the evaluation dataset.

        """
        pass

    @abstractmethod
    def prepare_for_deployment(self, *args, **kwargs):
        """Prepare the model data in JSON format for deployment on Jaqpot."""
        pass

    @staticmethod
    # @enforce_types
    def _model_data_as_json(
        actualModel: str,
        name: str,
        description: Union[str, None],
        model_type: str,
        visibility: str,
        independentFeatures: List[Feature],
        dependentFeatures: List[Feature],
        additional_model_params: dict,
        reliability: Optional[int] = None,
        pretrained: int = False,
        meta: Optional[dict] = None,
        organizations: Optional[List[Organization]] = None,
    ) -> dict:
        """Create a JSON representation of the model data for deployment.

        Args:
        ----
            actualModel (str): The actual model.
            name (str): The name of the model.
            description (Union[str, None]): The description of the model.
            model_type (str): The type of the model.
            visibility (str): The visibility of the model.
            independentFeatures (List[Feature]): List of independent features (inputs).
            dependentFeatures (List[Feature]): List of dependent features (targets).
            additional_model_params (dict): Additional model parameters.
            reliability (Optional[int]): The reliability of the model. Default is 0.
            pretrained (Optional[int]): Whether the model is pretrained. Default is False.
            meta (Optional[dict]): Additional metadata. Default is None.
            organizations (Optional[List[Organization]]): List of organizations. Default is None.

        Returns:
        -------
            dict: The JSON representation of the model data.

        """
        if not isinstance(name, str):
            msg = "'name' should be of type str"
            raise ValueError(msg)

        minLength_name = 3
        maxLength_name = 255
        if not minLength_name <= len(name) <= maxLength_name:
            msg = "Model 'name' length must be between 3 and 255 characters"
            raise ValueError(msg)

        description = "" if description is None else description
        if not isinstance(description, str):
            msg = "'description' should be of type str"
            raise ValueError(msg)

        if not isinstance(model_type, str):
            msg = "'model_type' should be of type str"
            raise ValueError(msg)
        if model_type not in TorchModelTrainer.get_subclass_model_types():
            msg = f"Invalid model_type: '{model_type}'. It must be one of {set(TorchModelTrainer.get_subclass_model_types())}."
            raise ValueError(msg)

        visibility_allowed_values = ["PUBLIC", "ORG_SHARED", "PRIVATE"]
        if visibility not in visibility_allowed_values:
            msg = f"Invalid visibility: {visibility}. Must be one of {visibility_allowed_values}"
            raise ValueError(msg)

        if not isinstance(independentFeatures, list):
            msg = "'independentFeatures' should be of type list"
            raise ValueError(msg)
        if any(not isinstance(feature, Feature) for feature in independentFeatures):
            msg = "'independentFeatures' should must only include elements of type Feature"
            raise ValueError(msg)

        if not isinstance(dependentFeatures, list):
            msg = "'dependentFeatures' should be of type list"
            raise ValueError(msg)
        if any(not isinstance(feature, Feature) for feature in independentFeatures):
            msg = (
                "'dependentFeatures' should must only include elements of type Feature"
            )
            raise ValueError(msg)

        if reliability is not None and not isinstance(reliability, int):
            msg = "'reliability' should be of type int"
            raise ValueError(msg)

        if not isinstance(pretrained, bool):
            msg = "'pretrained' should be of type bool"
            raise ValueError(msg)

        if meta is not None and not isinstance(meta, dict):
            msg = "'meta' should be of type dict"
            raise ValueError(msg)

        if organizations is not None and not isinstance(meta, list):
            msg = "'organizations' should be of type list"
            raise ValueError(msg)
        if organizations is not None and any(
            not isinstance(organization, Organization) for organization in organizations
        ):
            msg = (
                "'organizations' should must only include elements of type Organization"
            )
            raise ValueError(msg)

        if set(feature.name for feature in independentFeatures) & set(
            feature.name for feature in dependentFeatures
        ):
            msg = "There are input and output variables that might have the same naming"
            raise ValueError(msg)

        independent_feature_names = set(feature.name for feature in independentFeatures)
        dependent_feature_names = set(feature.name for feature in dependentFeatures)
        overlapping_features = independent_feature_names & dependent_feature_names
        if overlapping_features:
            msg = f"Input and output variables have the same name(s): {overlapping_features}"
            raise ValueError(msg)

        if meta is None:
            meta = dict()

        if organizations is None:
            organizations = []

        libraries = [
            Library(package_name, package_version)
            for package_name, package_version in get_installed_packages().items()
        ]

        jaqpotpyVersion = "1.0.0"
        # jaqpotpyVersion = str(jaqpotpy.__version__)

        data = {
            "meta": meta,
            "name": name,
            "description": description,
            "type": f"TORCH-{model_type}",
            "jaqpotpyVersion": jaqpotpyVersion,
            "libraries": [library.to_json() for library in libraries],
            "dependentFeatures": [feature.to_json() for feature in dependentFeatures],
            "independentFeatures": [
                feature.to_json() for feature in independentFeatures
            ],
            "organizations": organizations,
            "visibility": visibility,
            "reliability": reliability,
            "pretrained": pretrained,
            "actualModel": actualModel,
            "additional_model_params": additional_model_params,
        }

        return data

    def set_input_feature_meta_for_deployment(self, feature_name: str, meta: dict):
        """Set the meta attribute for an input feature in the JSON data for deployment.

        Args:
        ----
            feature_name (str): The name of the input feature.
            meta (dict): The meta data to set for the input feature.

        """
        if not isinstance(feature_name, str):
            msg = "'feature_name' should be of type str"
            raise ValueError(msg)
        if not isinstance(meta, dict):
            msg = "'meta' should be of type dict"
            raise ValueError(msg)
        self._set_feature_attr_for_deployment(
            feature_name, "meta", dict(meta), is_input_feature=True
        )

    def set_output_feature_meta_for_deployment(self, feature_name: str, meta: dict):
        """Set the meta attribute for an output feature in the JSON data for deployment.

        Args:
        ----
            feature_name (str): The name of the output feature.
            meta (dict): The meta data to set for the output feature.

        """
        if not isinstance(feature_name, str):
            msg = "'feature_name' should be of type str"
            raise ValueError(msg)
        if not isinstance(meta, dict):
            msg = "'meta' should be of type dict"
            raise ValueError(msg)
        self._set_feature_attr_for_deployment(
            feature_name, "meta", dict(meta), is_input_feature=False
        )

    def set_input_feature_description_for_deployment(
        self, feature_name: str, description: str
    ):
        """Set the description attribute for an input feature in the JSON data for deployment.

        Args:
        ----
            feature_name (str): The name of the input feature.
            description (str): The description to set for the input feature.

        """
        if not isinstance(feature_name, str):
            msg = "'feature_name' should be of type str"
            raise ValueError(msg)
        if not isinstance(description, str):
            msg = "'description' should be of type str"
            raise ValueError(msg)
        self._set_feature_attr_for_deployment(
            feature_name, "description", str(description), is_input_feature=True
        )

    def set_output_feature_description_for_deployment(
        self, feature_name: str, description: str
    ):
        """Set the description attribute for an output feature in the JSON data for deployment.

        Args:
        ----
            feature_name (str): The name of the output feature.
            description (str): The description to set for the output feature.

        """
        if not isinstance(feature_name, str):
            msg = "'feature_name' should be of type str"
            raise ValueError(msg)
        if not isinstance(description, str):
            msg = "'description' should be of type str"
            raise ValueError(msg)
        self._set_feature_attr_for_deployment(
            feature_name, "description", str(description), is_input_feature=False
        )

    def _set_feature_attr_for_deployment(
        self,
        feature_name,
        attr_name,
        attr_value,
        is_input_feature=True,
    ):
        """Set an attribute for a feature in the JSON data for deployment.

        Args:
        ----
            feature_name (str): The name of the feature.
            attr_name (str): The name of the attribute.
            attr_value: The value of the attribute.
            is_input_feature (bool, optional): Whether the feature is an input feature. Default is True.

        """
        if self.json_data_for_deployment is None:
            msg = "No JSON data for deployment to set feature attributes. You may need to call 'prepare_data_for_deployment' first."
            raise RuntimeError(msg)

        feature_dependency = (
            "independentFeatures" if is_input_feature else "dependentFeatures"
        )

        for i, feature in enumerate(self.json_data_for_deployment[feature_dependency]):
            if feature_name == feature["name"]:
                self.json_data_for_deployment[feature_dependency][i][attr_name] = (
                    attr_value
                )
                return

        raise ValueError(f"No feature named '{feature_name}'")

    def deploy_model(self, token):
        """Deploy the model to Jaqpot.

        Args:
        ----
            token (str): The authorization token.

        Returns:
        -------
            Response: The response from the Jaqpot server.

        """
        if self.json_data_for_deployment is None:
            no_json_err_msg = (
                "There are no JSON data to deploy. Note that prepare_for_deployment() must be called first to prepare the JSON to be deployed. "
                f"Please check the function documentation for the specific arguments needed for the current class '{self.__class__.__name__}'."
            )
            raise RuntimeError(no_json_err_msg)

        # TODO: Replace with the appropriate headers
        headers = {
            "Authorization": f"Bearer {token}",
            "Content-Type": "application/json",
        }

        # TODO: Replace with the appropriate url for model uploading
        # url = "http://localhost:8006/api/v1/models/upload/" # for jaqpotpy-torch-inference repo
        url = "http://localhost:8002/upload/"  # for jaqpotpy-inference repo

        # TODO: Add whatever else is needed in the POST request
        response = requests.post(
            url, json=self.json_data_for_deployment, headers=headers
        )

        return response



"""Author: Ioannis Pitoskas (jpitoskas@gmail.com)"""


class Library:
    def __init__(self, name: str, version: str):
        self.name = name
        self.version = version

    def to_json(self):
        library_dict = {
            "name": self.name,
            "version": self.version,
        }
        return library_dict

    def __str__(self):
        return self.to_json()

    def __repr__(self):
        return f"{self.__class__.__name__}({self.to_json()})"



import numpy as np

from jaqpotpy.utils.types import RDKitMol
from jaqpotpy.descriptors.base_classes import MolecularFeaturizer
from rdkit.Chem.AllChem import GetMACCSKeysFingerprint
from rdkit.Chem import DataStructs


class MACCSKeysFingerprint(MolecularFeaturizer):
    """MACCS Keys Fingerprint.
    The MACCS (Molecular ACCess System) keys are one of the most commonly used structural keys.
    Please confirm the details in [1]_, [2]_.

    Examples:
    --------
    >>> import jaqpotpy as jp
    >>> smiles = 'CC(=O)OC1=CC=CC=C1C(=O)O'
    >>> featurizer = jp.descriptors.molecular.MACCSKeysFingerprint()
    >>> features = featurizer.featurize([smiles])
    >>> type(features[0])
    <class 'numpy.ndarray'>
    >>> features[0].shape
    (167,)

    References:
    ----------
    .. [1] Durant, Joseph L., et al. "Reoptimization of MDL keys for use in drug discovery."
       Journal of chemical information and computer sciences 42.6 (2002): 1273-1280.
    .. [2] https://github.com/rdkit/rdkit/blob/master/rdkit/Chem/MACCSkeys.py

    Note:
    ----
    This class requires RDKit to be installed.

    """

    @property
    def __name__(self):
        return "MACCSKeysFingerprint"

    def __getitem__(self):
        return self

    def __init__(self):
        """Initialize this featurizer."""
        self.calculator = None

    def _featurize(self, datapoint: RDKitMol, **kwargs) -> np.ndarray:
        """Calculate MACCS keys fingerprint.

        Parameters
        ----------
        datapoint: rdkit.Chem.rdchem.Mol
          RDKit Mol object

        Returns
        -------
        np.ndarray
          1D array of RDKit descriptors for `mol`. The length is 167.

        """
        if "mol" in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )

        if self.calculator is None:
            try:
                self.calculator = GetMACCSKeysFingerprint
            except ModuleNotFoundError:
                raise ImportError("This class requires RDKit to be installed.")

        fp = self.calculator(datapoint)
        array = np.zeros((0,), dtype=np.int8)
        DataStructs.ConvertToNumpyArray(fp, array)
        return np.asarray(array)

    def _get_column_names(self, **kwargs) -> list:
        descriptors = []
        for i in range(167):
            descriptors.append("f" + str(i))
        return descriptors

    def _featurize_dataframe(self, datapoint: RDKitMol, **kwargs) -> np.ndarray:
        return self._featurize(datapoint, **kwargs)



from .graph_attention_network import GraphAttentionNetwork
from .graph_convolutional_network import GraphConvolutionalNetwork
from .graph_sage_network import GraphSAGENetwork
from .graph_transformer_network import GraphTransformerNetwork

from .fully_connected_network import FullyConnectedNetwork

from .graph_attention_network import GraphAttentionNetworkWithExternal
from .graph_convolutional_network import GraphConvolutionalNetworkWithExternal
from .graph_sage_network import GraphSAGENetworkWithExternal
from .graph_transformer_network import GraphTransformerNetworkWithExternal


from compress_pickle import dumps
from base64 import b64encode


def model_to_b64encoding(actualModel):
    p_mod = dumps(actualModel, compression="bz2")
    raw = b64encode(p_mod)
    model_b64encoded = raw.decode("utf-8")
    return model_b64encoded



from pydantic import parse_obj_as
from typing import Union, List
from jaqpotpy.entities.chemical_models import (
    ElementModel,
    IonicRadiiModel,
    IsotopesModel,
    IonizationEnergiesModel,
    OxidationStatesModel,
    ScreeningConstantsModel,
)
from jaqpotpy.helpers.periodic_table.base_classes import PeriodicTable


class Element(PeriodicTable):
    def __init__(self, element: Union[str, int]):
        super().__init__()
        if isinstance(element, str):
            if len(element) < 3:
                res = self._find_one(
                    'SELECT * FROM elements WHERE symbol="{}"'.format(element)
                )
            else:
                res = self._find_one(
                    'SELECT * FROM elements WHERE name="{}"'.format(element)
                )
        else:
            res = self._find_one(
                'SELECT * FROM elements WHERE atomic_number="{}"'.format(element)
            )

        try:
            g = self._find_one(
                "SELECT name, symbol FROM groups WHERE group_id={}".format(
                    res["group_id"]
                )
            )
        except:
            g = {"name": None, "symbol": None}
        try:
            s = self._find_one(
                "SELECT name,color FROM series WHERE id={}".format(res["series_id"])
            )
        except:
            s = {"name": None, "color": None}

        res["group_name"] = g["name"]
        res["group_symbol"] = g["symbol"]
        res["series_name"] = s["name"]
        res["series_color"] = s["color"]

        self._element = parse_obj_as(
            ElementModel, {k: None if not (v == v) else v for k, v in res.items()}
        )

    @property
    def parameters(self) -> ElementModel:
        return self._element

    @property
    def ionic_radii(self) -> List[IonicRadiiModel]:
        res = self._find_many(
            "SELECT * FROM ionicradii WHERE atomic_number={}".format(
                self._element.atomic_number
            )
        )
        return [
            parse_obj_as(
                IonicRadiiModel,
                {k: None if not (v == v) else v for k, v in rad.items()},
            )
            for rad in res
        ]

    @property
    def ionization_energies(self) -> List[IonizationEnergiesModel]:
        res = self._find_many(
            "SELECT * FROM ionizationenergies WHERE atomic_number={}".format(
                self._element.atomic_number
            )
        )
        return [
            parse_obj_as(
                IonizationEnergiesModel,
                {k: None if not (v == v) else v for k, v in rad.items()},
            )
            for rad in res
        ]

    @property
    def isotopes(self) -> List[IsotopesModel]:
        res = self._find_many(
            "SELECT * FROM isotopes WHERE atomic_number={}".format(
                self._element.atomic_number
            )
        )
        return [
            parse_obj_as(
                IsotopesModel, {k: None if not (v == v) else v for k, v in rad.items()}
            )
            for rad in res
        ]

    @property
    def oxidation_states(self) -> List[OxidationStatesModel]:
        res = self._find_many(
            "SELECT * FROM oxidationstates WHERE atomic_number={}".format(
                self._element.atomic_number
            )
        )
        return [
            parse_obj_as(
                OxidationStatesModel,
                {k: None if not (v == v) else v for k, v in rad.items()},
            )
            for rad in res
        ]

    @property
    def screeningconstants(self) -> List[ScreeningConstantsModel]:
        res = self._find_many(
            "SELECT * FROM screeningconstants WHERE atomic_number={}".format(
                self._element.atomic_number
            )
        )
        return [
            parse_obj_as(
                ScreeningConstantsModel,
                {k: None if not (v == v) else v for k, v in rad.items()},
            )
            for rad in res
        ]



"""Author: Ioannis Pitoskas (jpitoskas@gmail.com)"""

from . import MulticlassModelTrainer
import torch
import io
import base64
import pickle
from jaqpotpy.schemas import Feature
from typing import Optional
import inspect


class MulticlassGraphModelTrainer(MulticlassModelTrainer):
    """Trainer class for Multiclass Classification using Graph Neural Networks for SMILES and external features."""

    MODEL_TYPE = "multiclass-graph-model"
    """'multiclass-graph-model'"""

    @classmethod
    def get_model_type(cls):
        return cls.MODEL_TYPE

    def __init__(
        self,
        model,
        n_epochs,
        optimizer,
        loss_fn,
        scheduler=None,
        device="cpu",
        use_tqdm=True,
        log_enabled=True,
        log_filepath=None,
    ):
        """The MulticlassGraphModelTrainer constructor.

        Args:
        ----
            model (torch.nn.Module): The torch model to be trained.
            n_epochs (int): Number of training epochs.
            optimizer (torch.optim.Optimizer): The optimizer used for training the model.
            loss_fn (torch.nn.Module): The loss function used for training.
            scheduler (torch.optim.lr_scheduler.LRScheduler): The scheduler used for adjusting the learning rate during training. Default is None.
            device (str, optional): The device on which to train the model. Default is 'cpu'.
            use_tqdm (bool, optional): Whether to use tqdm for progress bars. Default is True.
            log_enabled (bool, optional): Whether logging is enabled. Default is True.
            log_filepath (str or None, optional): Path to the log file. If None, logging is not saved to a file. Default is None.

        Example:
        -------
        ```
        >>> import torch
        >>> from jaqpotpy.jaqpotpy_torch.models import GraphAttentionNetwork
        >>> from jaqpotpy.jaqpotpy_torch.trainers import MulticlassGraphModelTrainer
        >>>
        >>> num_classes = 10
        >>> model = GraphAttentionNetwork(input_dim=10,
        ...                               hidden_dims=[32, 32]
        ...                               edge_dim=5,
        ...                               output_dim=num_classes)
        >>> optimizer = torch.optim.Adam(model.parameters(), lr=0.0005)
        >>> loss_fn = torch.nn.CrossEntropyLoss()
        >>>
        >>> trainer = MulticlassGraphModelTrainer(model, n_epochs=50, optimizer=optimizer, loss_fn=loss_fn)
        ```

        """
        super().__init__(
            model=model,
            n_epochs=n_epochs,
            optimizer=optimizer,
            loss_fn=loss_fn,
            scheduler=scheduler,
            device=device,
            use_tqdm=use_tqdm,
            log_enabled=log_enabled,
            log_filepath=log_filepath,
        )

    def get_model_kwargs(self, data):
        """Fetch the model's keyword arguments.

        Args:
        ----
            data (torch_geometric.data.Data): Data object returned as returned by the Dataloader

        Returns:
        -------
            dict: The required model kwargs. Set of keywords: {'*x*', '*edge_index*', '*batch*', '*edge_attr*'}. Note that '*edge_attr*' is only present if the model supports edge features.

        """
        kwargs = {}

        kwargs["x"] = data.x
        kwargs["edge_index"] = data.edge_index
        kwargs["batch"] = data.batch

        if "edge_attr" in inspect.signature(self.model.forward).parameters:
            kwargs["edge_attr"] = data.edge_attr

        return kwargs

    def prepare_for_deployment(
        self,
        featurizer,
        endpoint_name: str,
        name: str,
        description: Optional[str] = None,
        visibility: str = "PUBLIC",
        reliability: Optional[int] = None,
        pretrained: bool = False,
        meta: dict = dict(),
    ):
        """Args:
        ----
            featurizer (object): The featurizer used to transform the SMILES to graph representations before training the model.
            endpoint_name (str): The name of the endpoint for the deployed model.
            name (str): The name to be assigned to the deployed model.
            description (str, optional): A description for the model to be deployed. Default is None.
            visibility (str, optional): Visibility of the deployed model. Can be 'PUBLIC', 'PRIVATE' or 'ORG_SHARED'. Default is 'PUBLIC'.
            reliability (int, optional): The models reliability. Default is None.
            pretrained (bool, optional): Indicates if the model is pretrained. Default is False.
            meta (dict, optional): Additional metadata for the model. Default is an empty dictionary.

        Returns
        -------
            dict: The data to be sent to the API of Jaqpot in JSON format.
                  Note that in this case, the '*additional_model_params*' key contains a nested dictionary with they keys: {'*featurizer*'}.

        """
        self.model = self.model.cpu()
        model_scripted = torch.jit.script(self.model)
        model_buffer = io.BytesIO()
        torch.jit.save(model_scripted, model_buffer)
        model_buffer.seek(0)
        model_scripted_base64 = base64.b64encode(model_buffer.getvalue()).decode(
            "utf-8"
        )

        featurizer_buffer = io.BytesIO()
        pickle.dump(featurizer, featurizer_buffer)
        featurizer_buffer.seek(0)
        featurizer_pickle_base64 = base64.b64encode(
            featurizer_buffer.getvalue()
        ).decode("utf-8")

        additional_model_params = {"featurizer": featurizer_pickle_base64}

        independentFeatures = [
            Feature(
                name="SMILES",
                featureDependency="INDEPENDENT",
                possibleValues=[],
                featureType="SMILES",
            )
        ]

        num_classes = self.model.output_dim
        dependentFeatures = [
            Feature(
                name=endpoint_name,
                featureDependency="DEPENDENT",
                possibleValues=[str(i) for i in range(num_classes)],
                featureType="CATEGORICAL",
            )
        ]

        self.json_data_for_deployment = self._model_data_as_json(
            actualModel=model_scripted_base64,
            name=name,
            description=description,
            model_type=self.get_model_type(),
            visibility=visibility,
            independentFeatures=independentFeatures,
            dependentFeatures=dependentFeatures,
            additional_model_params=additional_model_params,
            reliability=reliability,
            pretrained=pretrained,
            meta=meta,
        )

        return self.json_data_for_deployment



from jaqpotpy.descriptors.base_classes import MaterialFeaturizer
from pymatgen.util.typing import CompositionLike, Composition
import numpy as np
import pandas as pd


class ElementNet(MaterialFeaturizer):
    """Create a fixed vector of length 118, containing rwa, fractional
    element compostitions in a compound.

    References:
    ----------
    Jha, D., Ward, L., Paul, A. et al. Sci Rep 8, 17593 (2018).
       https://doi.org/10.1038/s41598-018-35934-y

    Examples:
    --------
    >>> import jaqpotpy as jt
    >>> comp = "Fe2O3"
    >>> featurizer = jt.descriptors.material.ElementNet()
    >>> features = featurizer.featurize([comp])
    >>> type(features[0])
    <class 'numpy.ndarray'>
    >>> features[0].shape
    (118,)

    Note:
    ----
    This class requires matminer and Pymatgen to be installed.
    `NaN` feature values are automatically converted to 0 by this featurizer.

    """

    def __init__(self):
        """Create a fixed vector of length 118, containing rwa, fractional
        element compostitions in a compound.

        References
        ----------
          Jha, D., Ward, L., Paul, A. et al. Sci Rep 8, 17593 (2018).
             https://doi.org/10.1038/s41598-018-35934-y

        """
        self.elements = [
            "H",
            "He",
            "Li",
            "Be",
            "B",
            "C",
            "N",
            "O",
            "F",
            "Ne",
            "Na",
            "Mg",
            "Al",
            "Si",
            "P",
            "S",
            "Cl",
            "Ar",
            "K",
            "Ca",
            "Sc",
            "Ti",
            "V",
            "Cr",
            "Mn",
            "Fe",
            "Co",
            "Ni",
            "Cu",
            "Zn",
            "Ga",
            "Ge",
            "As",
            "Se",
            "Br",
            "Kr",
            "Rb",
            "Sr",
            "Y",
            "Zr",
            "Nb",
            "Mo",
            "Tc",
            "Ru",
            "Rh",
            "Pd",
            "Ag",
            "Cd",
            "In",
            "Sn",
            "Sb",
            "Te",
            "I",
            "Xe",
            "Cs",
            "Ba",
            "La",
            "Ce",
            "Pr",
            "Nd",
            "Pm",
            "Sm",
            "Eu",
            "Gd",
            "Tb",
            "Dy",
            "Ho",
            "Er",
            "Tm",
            "Yb",
            "Lu",
            "Hf",
            "Ta",
            "W",
            "Re",
            "Os",
            "Ir",
            "Pt",
            "Au",
            "Hg",
            "Tl",
            "Pb",
            "Bi",
            "Po",
            "At",
            "Rn",
            "Fr",
            "Ra",
            "Ac",
            "Th",
            "Pa",
            "U",
            "Np",
            "Pu",
            "Am",
            "Cm",
            "Bk",
            "Cf",
            "Es",
            "Fm",
            "Md",
            "No",
            "Lr",
            "Rf",
            "Db",
            "Sg",
            "Bh",
            "Hs",
            "Mt",
            "Ds",
            "Rg",
            "Cn",
            "Nh",
            "Fl",
            "Mc",
            "Lv",
            "Ts",
            "Og",
        ]
        return

    @property
    def __name__(self):
        return "ElementNet"

    def __getitem__(self):
        return self

    def _featurize(self, datapoint: CompositionLike, **kwargs) -> np.ndarray:
        """Calculate 118 dimensional vector containing fractional
        compositions of each element in the compound.

        Parameters
        ----------
        datapoint: pymatgen.utils.typing.CompositionLike object
          Composition or an object that can be used to creaet a composition

        Returns
        -------
        feats: np.ndarray
          Vector containing the fractional compositions of each element.

        """
        if not isinstance(datapoint, Composition):
            datapoint = Composition(datapoint)
        fractions = datapoint.fractional_composition.get_el_amt_dict()
        return np.array([fractions[e] if e in fractions else 0 for e in self.elements])

    def _featurize_dataframe(
        self, datapoint: CompositionLike, **kwargs
    ) -> pd.DataFrame:
        """Calculate 118 dimensional vector containing fractional
        compositions of each element in the compound.

        Parameters
        ----------
        datapoint: pymatgen.utils.typing.CompositionLike object
          Composition or an object that can be used to create a composition

        Returns
        -------
        feats: pd.Dataframe
          Dataframe of the vector containing the fractional compositions of each element.

        """
        data = self._featurize(datapoint)
        return pd.DataFrame([data], columns=self.elements, index=[0])



import inspect
import logging
import numpy as np
from typing import Any, List, Iterable, Optional, Tuple, Union, cast
import pandas as pd
from tqdm import tqdm
from jaqpotpy.cfg import config

logger = logging.getLogger(__name__)
_print_threshold = 10


class Featurizer(object):
    """Abstract class for calculating a set of features for a datapoint.
    This class is abstract and cannot be invoked directly. You'll
    likely only interact with this class if you're a developer. In
    that case, you might want to make a child class which
    implements the `_featurize` method for calculating features for
    a single datapoints if you'd like to make a featurizer for a
    new datatype.

    References
    ----------
    .. [1] Ramsundar-et-al-2019,. "Deep Learning for the Life Sciences."
       O'Reilly Media (2019):.
    .. [2] https://github.com/deepchem/deepchem

    """

    def featurize(
        self, datapoints: Iterable[Any], log_every_n: int = 1000, **kwargs
    ) -> np.ndarray:
        """Calculate features for datapoints.

        Parameters
        ----------
        datapoints: Iterable[Any]
          A sequence of objects that you'd like to featurize. Subclassses of
          `Featurizer` should instantiate the `_featurize` method that featurizes
          objects in the sequence.
        log_every_n: int, default 1000
          Logs featurization progress every `log_every_n` steps.

        Returns
        -------
        np.ndarray
          A numpy array containing a featurized representation of `datapoints`.

        """
        datapoints = list(datapoints)
        features = []
        if config.verbose is False:
            disable_tq = True
        else:
            disable_tq = False
        for i, point in enumerate(
            tqdm(datapoints, desc="Creating descriptors", disable=disable_tq)
        ):
            try:
                features.append(self._featurize(point, **kwargs))
            except:
                if config.verbose is True:
                    logger.warning(
                        "Failed to featurize datapoint %d. Appending empty array"
                    )
                features.append(np.array([]))

        return np.asarray(features)

    def __call__(self, datapoints: Iterable[Any], **kwargs):
        """Calculate features for datapoints.
        `**kwargs` will get passed directly to `Featurizer.featurize`

        Parameters
        ----------
        datapoints: Iterable[Any]
          Any blob of data you like. Subclasss should instantiate this.

        """
        return self.featurize(datapoints, **kwargs)

    def _featurize(self, datapoint: Any, **kwargs):
        """Calculate features for a single datapoint.

        Parameters
        ----------
        datapoint: Any
          Any blob of data you like. Subclass should instantiate this.

        """
        raise NotImplementedError("Featurizer is not defined.")

    def _featurize_dataframe(self, datapoint: Any, **kwargs):
        """Calculate features for a single datapoint.

        Parameters
        ----------
        datapoint: Any
          Any blob of data you like. Subclass should instantiate this.

        """
        raise NotImplementedError("Featurizer is not defined.")

    def __repr__(self) -> str:
        """Convert self to repr representation.

        Returns
        -------
        str
          The string represents the class.

        Examples
        --------
        >>> import jaqpotpy as jt
        >>> jt.descriptors.CircularFingerprint(size=1024, radius=4)
        CircularFingerprint[radius=4, size=1024, chiral=False, bonds=True, features=False, sparse=False, smiles=False]
        >>> jt.descriptors.CGCNNFeaturizer()
        CGCNNFeaturizer[radius=8.0, max_neighbors=12, step=0.2]

        """
        args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
        args_names = [arg for arg in args_spec.args if arg != "self"]
        args_info = ""
        for arg_name in args_names:
            value = self.__dict__[arg_name]
            # for str
            if isinstance(value, str):
                value = "'" + value + "'"
            # for list
            if isinstance(value, list):
                threshold = get_print_threshold()
                value = np.array2string(np.array(value), threshold=threshold)
            args_info += arg_name + "=" + str(value) + ", "
        return self.__class__.__name__ + "[" + args_info[:-2] + "]"

    def __str__(self) -> str:
        """Convert self to str representation.

        Returns
        -------
        str
          The string represents the class.

        Examples
        --------
        >>> import jaqpotpy as jt
        >>> jt.descriptors.CircularFingerprint(size=1024, radius=4)
        CircularFingerprint[radius=4, size=1024, chiral=False, bonds=True, features=False, sparse=False, smiles=False]
        >>> jt.descriptors.CGCNNFeaturizer()
        CGCNNFeaturizer[radius=8.0, max_neighbors=12, step=0.2]

        """
        args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
        args_names = [arg for arg in args_spec.args if arg != "self"]
        args_num = len(args_names)
        args_default_values = [None for _ in range(args_num)]
        if args_spec.defaults is not None:
            defaults = list(args_spec.defaults)
            args_default_values[-len(defaults) :] = defaults

        override_args_info = ""
        for arg_name, default in zip(args_names, args_default_values):
            if arg_name in self.__dict__:
                arg_value = self.__dict__[arg_name]
                # validation
                # skip list
                if isinstance(arg_value, list):
                    continue
                if isinstance(arg_value, str):
                    # skip path string
                    if "\\/." in arg_value or "/" in arg_value or "." in arg_value:
                        continue
                # main logic
                if default != arg_value:
                    override_args_info += "_" + arg_name + "_" + str(arg_value)
        return self.__class__.__name__ + override_args_info


class ComplexFeaturizer(Featurizer):
    """ "
    Abstract class for calculating features for mol/protein complexes.
    """

    def featurize(
        self,
        datapoints: Optional[Iterable[Tuple[str, str]]] = None,
        log_every_n: int = 100,
        **kwargs,
    ) -> np.ndarray:
        """Calculate features for mol/protein complexes.

        Parameters
        ----------
        datapoints: Iterable[Tuple[str, str]]
          List of filenames (PDB, SDF, etc.) for ligand molecules and proteins.
          Each element should be a tuple of the form (ligand_filename,
          protein_filename).

        Returns
        -------
        features: np.ndarray
          Array of features

        """
        if "complexes" in kwargs:
            datapoints = kwargs.get("complexes")
            raise DeprecationWarning(
                'Complexes is being phased out as a parameter, please pass "datapoints" instead.'
            )
        if not isinstance(datapoints, Iterable):
            datapoints = [cast(Tuple[str, str], datapoints)]
        features, failures, successes = [], [], []
        for idx, point in enumerate(datapoints):
            if idx % log_every_n == 0:
                logger.info("Featurizing datapoint %i" % idx)
            try:
                features.append(self._featurize(point, **kwargs))
                successes.append(idx)
            except:
                logger.warning(
                    "Failed to featurize datapoint %i. Appending empty array." % idx
                )
                features.append(np.zeros(1))
                failures.append(idx)

        # Find a successful featurization
        try:
            i = np.argmax([f.shape[0] for f in features])
            dtype = features[i].dtype
            shape = features[i].shape
            dummy_array = np.zeros(shape, dtype=dtype)
        except AttributeError:
            dummy_array = features[successes[0]]

        # Replace failed featurizations with appropriate array
        for idx in failures:
            features[idx] = dummy_array

        return np.asarray(features)

    def _featurize(self, datapoint: Optional[Tuple[str, str]] = None, **kwargs):
        """Calculate features for single mol/protein complex.

        Parameters
        ----------
        complex: Tuple[str, str]
          Filenames for molecule and protein.

        """
        raise NotImplementedError("Featurizer is not defined.")


class DataframeFeaturizer(object):
    """Abstract class for calculating a set of features for a datapoint.
    This class is abstract and cannot be invoked directly. You'll
    likely only interact with this class if you're a developer. In
    that case, you might want to make a child class which
    implements the `_featurize` method for calculating features for
    a single datapoints if you'd like to make a featurizer for a
    new datatype.

    References
    ----------
    .. [1] Ramsundar-et-al-2019,. "Deep Learning for the Life Sciences."
       O'Reilly Media (2019):.
    .. [2] https://github.com/deepchem/deepchem

    """

    def featurize(
        self, datapoints: Iterable[Any], log_every_n: int = 1000, **kwargs
    ) -> Any:
        """Calculate features for datapoints.

        Parameters
        ----------
        datapoints: Iterable[Any]
          A sequence of objects that you'd like to featurize. Subclassses of
          `Featurizer` should instantiate the `_featurize` method that featurizes
          objects in the sequence.
        log_every_n: int, default 1000
          Logs featurization progress every `log_every_n` steps.

        Returns
        -------
        np.ndarray
          A numpy array containing a featurized representation of `datapoints`.

        """
        datapoints = list(datapoints)
        features = []
        if config.verbose is False:
            disable_tq = True
        else:
            disable_tq = False
        for i, point in enumerate(
            tqdm(datapoints, desc="Creating descriptors", disable=disable_tq)
        ):
            if i % log_every_n == 0:
                logger.info("Featurizing datapoint %i" % i)
            try:
                features.append(self._featurize(point, **kwargs))
            except:
                if config.verbose is True:
                    logger.warning(
                        "Failed to featurize datapoint %d. Appending empty array"
                    )
                features.append(np.array([]))

        return np.asarray(features)

    def __call__(self, datapoints: Iterable[Any], **kwargs):
        """Calculate features for datapoints.
        `**kwargs` will get passed directly to `Featurizer.featurize`

        Parameters
        ----------
        datapoints: Iterable[Any]
          Any blob of data you like. Subclasss should instantiate this.

        """
        return self.featurize(datapoints, **kwargs)

    def _featurize(self, datapoint: Any, **kwargs):
        """Calculate features for a single datapoint.

        Parameters
        ----------
        datapoint: Any
          Any blob of data you like. Subclass should instantiate this.

        """
        raise NotImplementedError("Featurizer is not defined.")

    def __repr__(self) -> str:
        """Convert self to repr representation.

        Returns
        -------
        str
          The string represents the class.

        Examples
        --------
        >>> import jaqpotpy as jt
        >>> jt.descriptors.CircularFingerprint(size=1024, radius=4)
        CircularFingerprint[radius=4, size=1024, chiral=False, bonds=True, features=False, sparse=False, smiles=False]
        >>> jt.descriptors.CGCNNFeaturizer()
        CGCNNFeaturizer[radius=8.0, max_neighbors=12, step=0.2]

        """
        args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
        args_names = [arg for arg in args_spec.args if arg != "self"]
        args_info = ""
        for arg_name in args_names:
            value = self.__dict__[arg_name]
            # for str
            if isinstance(value, str):
                value = "'" + value + "'"
            # for list
            if isinstance(value, list):
                threshold = get_print_threshold()
                value = np.array2string(np.array(value), threshold=threshold)
            args_info += arg_name + "=" + str(value) + ", "
        return self.__class__.__name__ + "[" + args_info[:-2] + "]"

    def __str__(self) -> str:
        """Convert self to str representation.

        Returns
        -------
        str
          The string represents the class.

        Examples
        --------
        >>> import jaqpotpy as jt
        >>> jt.descriptors.CircularFingerprint(size=1024, radius=4)
        CircularFingerprint[radius=4, size=1024, chiral=False, bonds=True, features=False, sparse=False, smiles=False]
        >>> jt.descriptors.CGCNNFeaturizer()
        CGCNNFeaturizer[radius=8.0, max_neighbors=12, step=0.2]

        """
        args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
        args_names = [arg for arg in args_spec.args if arg != "self"]
        args_num = len(args_names)
        args_default_values = [None for _ in range(args_num)]
        if args_spec.defaults is not None:
            defaults = list(args_spec.defaults)
            args_default_values[-len(defaults) :] = defaults

        override_args_info = ""
        for arg_name, default in zip(args_names, args_default_values):
            if arg_name in self.__dict__:
                arg_value = self.__dict__[arg_name]
                # validation
                # skip list
                if isinstance(arg_value, list):
                    continue
                if isinstance(arg_value, str):
                    # skip path string
                    if "\\/." in arg_value or "/" in arg_value or "." in arg_value:
                        continue
                # main logic
                if default != arg_value:
                    override_args_info += "_" + arg_name + "_" + str(arg_value)
        return self.__class__.__name__ + override_args_info


class MolecularFeaturizer(Featurizer):
    """Abstract class for calculating a set of features for a
    molecule.
    The defining feature of a `MolecularFeaturizer` is that it
    uses SMILES strings and RDKit molecule objects to represent
    small molecules. All other featurizers which are subclasses of
    this class should plan to process input which comes as smiles
    strings or RDKit molecules.
    Child classes need to implement the _featurize method for
    calculating features for a single molecule.

    .. [1] Ramsundar-et-al-2019,. "Deep Learning for the Life Sciences."
       O'Reilly Media (2019):.
    .. [2] https://github.com/deepchem/deepchem

    Note:
    ----
    The subclasses of this class require RDKit to be installed.

    """

    def featurize(self, datapoints, log_every_n=1000, **kwargs) -> np.ndarray:
        """Calculate features for molecules.

        Parameters
        ----------
        datapoints: rdkit.Chem.rdchem.Mol / SMILES string / iterable
          RDKit Mol, or SMILES string or iterable sequence of RDKit mols/SMILES
          strings.
        log_every_n: int, default 1000
          Logging messages reported every `log_every_n` samples.

        Returns
        -------
        features: np.ndarray
          A numpy array containing a featurized representation of `datapoints`.

        """
        try:
            from rdkit import Chem
            from rdkit.Chem import rdmolfiles
            from rdkit.Chem import rdmolops
            from rdkit.Chem.rdchem import Mol
        except ModuleNotFoundError:
            raise ImportError("This class requires RDKit to be installed.")

        if "molecules" in kwargs:
            datapoints = kwargs.get("molecules")
            raise DeprecationWarning(
                'Molecules is being phased out as a parameter, please pass "datapoints" instead.'
            )

        # Special case handling of single molecule
        if isinstance(datapoints, str) or isinstance(datapoints, Mol):
            datapoints = [datapoints]
        else:
            # Convert iterables to list
            datapoints = list(datapoints)

        features: list = []
        if config.verbose is False:
            disable_tq = True
        else:
            disable_tq = False
        for i, mol in enumerate(
            tqdm(datapoints, desc="Creating descriptors", disable=disable_tq)
        ):
            if i % log_every_n == 0:
                logger.info("Featurizing datapoint %i" % i)

            try:
                if isinstance(mol, str):
                    # mol must be a RDKit Mol object, so parse a SMILES
                    mol = Chem.MolFromSmiles(mol)
                    # SMILES is unique, so set a canonical order of atoms
                    new_order = rdmolfiles.CanonicalRankAtoms(mol)
                    mol = rdmolops.RenumberAtoms(mol, new_order)

                features.append(self._featurize(mol, **kwargs))
            except Exception as e:
                if isinstance(mol, Chem.rdchem.Mol):
                    mol = Chem.MolToSmiles(mol)
                if config.verbose is True:
                    logger.warning(
                        "Failed to featurize datapoint %d, %s. Appending empty array",
                        i,
                        mol,
                    )
                    logger.warning("Exception message: {}".format(e))
                features.append(np.array([]))

        return np.asarray(features)

    def featurize_dataframe(self, datapoints, log_every_n=1000, **kwargs) -> Any:
        """Calculate features for molecules.

        Parameters
        ----------
        datapoints: rdkit.Chem.rdchem.Mol / SMILES string / iterable
          RDKit Mol, or SMILES string or iterable sequence of RDKit mols/SMILES
          strings.
        log_every_n: int, default 1000
          Logging messages reported every `log_every_n` samples.

        Returns
        -------
        features: pd.Dataframe()
          A pandas Dataframe containing a featurized representation of `datapoints`.

        """
        try:
            from rdkit import Chem
            from rdkit.Chem import rdmolfiles
            from rdkit.Chem import rdmolops
            from rdkit.Chem.rdchem import Mol
        except ModuleNotFoundError:
            raise ImportError("This class requires RDKit to be installed.")

        if "molecules" in kwargs:
            datapoints = kwargs.get("molecules")
            raise DeprecationWarning(
                'Molecules is being phased out as a parameter, please pass "datapoints" instead.'
            )

        # Special case handling of single molecule
        if isinstance(datapoints, str) or isinstance(datapoints, Mol):
            datapoints = [datapoints]
        else:
            # Convert iterables to list
            datapoints = list(datapoints)

        features: list = []
        if config.verbose is False:
            disable_tq = True
        else:
            disable_tq = False
        for i, mol in enumerate(
            tqdm(datapoints, desc="Creating descriptors", disable=disable_tq)
        ):
            try:
                if isinstance(mol, str):
                    # mol must be a RDKit Mol object, so parse a SMILES
                    mol = Chem.MolFromSmiles(mol)
                    # SMILES is unique, so set a canonical order of atoms
                    new_order = rdmolfiles.CanonicalRankAtoms(mol)
                    mol = rdmolops.RenumberAtoms(mol, new_order)
                    ar = self._featurize_dataframe(mol, **kwargs)
                else:
                    ar = self._featurize_dataframe(mol, **kwargs)
                features.append(ar)
            except Exception as e:
                if isinstance(mol, Chem.rdchem.Mol):
                    mol = Chem.MolToSmiles(mol)
                if config.verbose is True:
                    logger.warning(
                        "Failed to featurize datapoint %d, %s. Appending empty array",
                        i,
                        mol,
                    )
                    logger.warning("Exception message: {}".format(e))
                features.append(np.array([]))
                # features.append(self._featurize_dataframe(mol, **kwargs))
        columns = self._get_column_names()
        # features = np.array(features)
        if columns == ["Sequence"]:
            df = pd.DataFrame({"Sequence": features})
        elif columns == ["OneHotSequence"]:
            df = pd.DataFrame({"OneHotSequence": features})
        elif columns == ["SmilesImage"]:
            df = pd.DataFrame({"SmilesImage": features})
        elif columns == ["MACCSFingerprint"]:
            df = pd.DataFrame({"MACCSFingerprint": features})
        elif columns == ["MolGanGraphs"]:
            df = pd.DataFrame({"MolGanGraphs": features})
        elif columns == ["CoulombMatrixEig"]:
            df = pd.DataFrame({"CoulombMatrixEig": features})
        else:
            df = pd.DataFrame(features, columns=columns)
        return df


class ParserFeaturizer(Featurizer):
    """Abstract class for calculating a set of features for one or
    a set of files. The defining feature of a `ParserFeaturizer`
    is that it uses files to generate descriptors. All other
    featurizers which are subclasses of this class
    should plan to process input which comes as a parser of a file or
    a folder with files. Child classes need to implement the
    _featurize method for calculating features for a single instance.
    """

    def featurize(self, log_every_n=1000, **kwargs) -> np.ndarray:
        """Calculate features for materials.

        Parameters
        ----------
        material: string
          Either the path of a pdb file or the path to the folder with many pdb files.
        log_every_n: int, default 1000
          Logging messages reported every `log_every_n` samples.
        files: str
          Extention of the files - Type of representation of materials

        Returns
        -------
        features: np.ndarray
          A numpy array containing a featurized representation of `datapoints`.

        """
        features: list = []
        if config.verbose is False:
            disable_tq = True
        else:
            disable_tq = False
        for i, mat in enumerate(
            tqdm(self.parser.parse(), desc="Creating descriptors", disable=disable_tq)
        ):
            if i % log_every_n == 0:
                logger.info("Featurizing datapoint %i" % i)

            try:
                features.append(self._featurize(mat, **kwargs))

            except Exception as e:
                if config.verbose is True:
                    logger.warning(
                        "Failed to featurize datapoint %d, %s. Appending empty array",
                        i,
                        mat,
                    )
                    logger.warning("Exception message: {}".format(e))
                features.append(np.array([]))
        return np.asarray(features)

    def featurize_dataframe(self, log_every_n=1000, **kwargs) -> Any:
        """Calculate features for materials.

        Parameters
        ----------
        pdb: string
          Either the path of a pdb file or the path to the folder with many pdb files.
        log_every_n: int, default 1000
          Logging messages reported every `log_every_n` samples.
        files: str
          Extention of the files - Type of representation of materials.

        Returns
        -------
        features: pd.Dataframe()
          A pandas Dataframe containing a featurized representation of `datapoints`.

        """
        features = pd.DataFrame()
        if config.verbose is False:
            disable_tq = True
        else:
            disable_tq = False
        for i, mat in enumerate(
            tqdm(self.parser.parse(), desc="Creating descriptors", disable=disable_tq)
        ):
            if i % log_every_n == 0:
                logger.info("Featurizing datapoint %i" % i)

            try:
                features = (
                    pd.concat([features, self._featurize_dataframe(mat, **kwargs)])
                    .reset_index(drop=True)
                    .replace(np.nan, 0)
                )

            except Exception as e:
                if config.verbose is True:
                    logger.warning(
                        "Failed to featurize datapoint %d, %s. Appending empty array", i
                    )
                    logger.warning("Exception message: {}".format(e))
                features = pd.concat([features, pd.DataFrame()]).reset_index(drop=True)

        return features


class MaterialFeaturizer(Featurizer):
    """Abstract class for calculating a set of features for an
    inorganic crystal composition.
    The defining feature of a `MaterialFeaturizer` is that it
    operates on 3D crystal materials.
    Inorganic crystal compositions are represented by Pymatgen composition
    objects. Featurizers for inorganic crystal compositions that are
    subclasses of this class should plan to process input which comes as
    Pymatgen composition objects.
    This class is abstract and cannot be invoked directly. You'll
    likely only interact with this class if you're a developer. Child
    classes need to implement the _featurize method for calculating
    features for a single crystal composition.

    Note:
    ----
    Some subclasses of this class will require pymatgen and matminer to be
    installed.

    """

    def featurize(
        self, datapoints: Union[List, Iterable[str]], log_every_n: int = 1000, **kwargs
    ) -> np.ndarray:
        """Calculate features for crystal compositions.

        Parameters
        ----------
        datapoints: Iterable[str]
          Iterable sequence of composition strings, e.g. "MoS2".
        log_every_n: int, default 1000
          Logging messages reported every `log_every_n` samples.

        Returns
        -------
        features: np.ndarray
          A numpy array containing a featurized representation of
          `compositions`.

        """
        if not isinstance(datapoints, List):
            datapoints = [datapoints]

        if config.verbose is False:
            disable_tq = True
        else:
            disable_tq = False

        features = []
        for i, item in enumerate(
            tqdm(datapoints, desc="Creating descriptors", disable=disable_tq)
        ):
            if i % log_every_n == 0:
                logger.info("Featurizing datapoint %i" % i)

            try:
                features.append(self._featurize(item, **kwargs))
            except Exception as e:
                if config.verbose is True:
                    logger.warning(
                        "Failed to featurize datapoint %i. Appending empty array" % i, e
                    )
                features.append(np.array([]))

        return np.asarray(features)

    def featurize_dataframe(
        self, datapoints: Union[List, Iterable[str]], log_every_n=1000, **kwargs
    ) -> pd.DataFrame:
        """Calculate features for materials.

        Parameters
        ----------
        pdb: string
          Either the path of a pdb file or the path to the folder with many pdb files.
        log_every_n: int, default 1000
          Logging messages reported every `log_every_n` samples.
        files: str
          Extention of the files - Type of representation of materials.

        Returns
        -------
        features: pd.Dataframe()
          A pandas Dataframe containing a featurized representation of `datapoints`.

        """
        if not isinstance(datapoints, List):
            datapoints = [datapoints]

        if config.verbose is False:
            disable_tq = True
        else:
            disable_tq = False

        datapoints = list(datapoints)
        features = pd.DataFrame()
        for i, item in enumerate(
            tqdm(datapoints, desc="Creating descriptors", disable=disable_tq)
        ):
            if i % log_every_n == 0:
                logger.info("Featurizing datapoint %i" % i)

            try:
                features = pd.concat(
                    [features, self._featurize_dataframe(item, **kwargs)]
                ).reset_index(drop=True)

            except Exception as e:
                if config.verbose is True:
                    logger.warning(
                        "Failed to featurize datapoint %d, %s. Appending empty array", i
                    )
                    logger.warning("Exception message: {}".format(e))
                features = pd.concat([features, pd.DataFrame()]).reset_index(drop=True)

        return features


def get_print_threshold() -> int:
    """Return the printing threshold for datasets.
    The print threshold is the number of elements from ids/tasks to
    print when printing representations of `Dataset` objects.

    Returns
    -------
    threshold: int
      Number of elements that will be printed

    """
    return _print_threshold



import http.client as http_client
import json
import time
import webbrowser

import numpy as np
import pandas as pd
from keycloak import KeycloakOpenID

import jaqpotpy.api.dataset_api as data_api
import jaqpotpy.api.doa_api as doa_api
import jaqpotpy.api.feature_api as featapi
import jaqpotpy.api.models_api as models_api
import jaqpotpy.api.task_api as task_api
import jaqpotpy.doa.doa as jha
import jaqpotpy.helpers.dataset_deserializer as ds
from jaqpotpy.api.openapi.jaqpot_api_client.api.model import create_model
from jaqpotpy.api.openapi.jaqpot_api_client.models import Model
from jaqpotpy.api.openapi.jaqpot_api_client.models.model_visibility import ModelVisibility
from jaqpotpy.api.openapi.jaqpot_api_client.models.feature import Feature
from jaqpotpy.api.openapi.jaqpot_api_client.client import AuthenticatedClient
from jaqpotpy.api.model_to_b64encoding import model_to_b64encoding
from jaqpotpy.helpers.logging import init_logger
from jaqpotpy.helpers.serializer import JaqpotSerializer
from jaqpotpy.utils.url_utils import add_subdomain

ENCODING = "utf-8"


# import matplotlib
# matplotlib.use('TkAgg')
# import matplotlib.pyplot as plt


class Jaqpot:
    """Deploys sklearn models on Jaqpot.

    Extended description of function.

    Parameters
    ----------
    base_url : The url on which Jaqpot services are deployed

    """

    def __init__(self, base_url=None, create_logs=False):
        # logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.INFO)
        self.log = init_logger(
            __name__, testing_mode=False, output_log_file=create_logs
        )
        if base_url:
            self.base_url = base_url
        else:
            self.base_url = "https://appv2.jaqpot.org/"
        self.app_url = add_subdomain(self.base_url, "app")
        self.login_url = add_subdomain(self.base_url, "login")
        self.api_url = add_subdomain(self.base_url, "api")
        self.api_key = None
        self.user_id = None
        self.http_client = http_client

    def login(self):
        """Logins on Jaqpot."""
        try:
            # Configure Keycloak client
            keycloak_openid = KeycloakOpenID(
                server_url=self.login_url,
                client_id="jaqpot-client",
                realm_name="jaqpot",
            )

            # Generate the authorization URL
            redirect_uri = "urn:ietf:wg:oauth:2.0:oob"
            auth_url = keycloak_openid.auth_url(
                redirect_uri=redirect_uri,
                scope="openid email profile",
                state="random_state_value",
            )

            print(f"Open this URL in your browser and log in:\n{auth_url}")

            # Automatically open the browser (optional)
            webbrowser.open(auth_url)

            code = input("Enter the authorization code you received: ")

            # Exchange the code for an access token
            token_response = keycloak_openid.token(
                grant_type="authorization_code", code=code, redirect_uri=redirect_uri
            )

            access_token = token_response["access_token"]
            self.api_key = access_token
        except Exception:
            self.log.error("Could not login to jaqpot")

    def set_api_key(self, api_key):
        """Set's api key for authentication on Jaqpot.

        Parameters
        ----------
        api_key : api_key can be retireved from the application after logged in

        """
        self.api_key = api_key
        self.log.info("api key is set")

    def deploy_sklearn_model(self, model, name, description, visibility):
        """ "
        Deploy sklearn models on Jaqpot.
        :param model:
        :param name:
        :param description:
        :param visibility:
        :return:
        """
        auth_client = AuthenticatedClient(base_url=self.api_url, token=self.api_key)
        actual_model = model_to_b64encoding(model.copy())
        body_model = Model(
            name=name,
            type=model.type,
            jaqpotpy_version=model.jaqpotpy_version,
            libraries=model.libraries,
            dependent_features=[
                Feature(
                    key=feature_i["key"],
                    name=feature_i["name"],
                    feature_type=feature_i["featureType"],
                )
                for feature_i in model.dependentFeatures
            ],
            independent_features=[
                Feature(
                    key=feature_i["key"],
                    name=feature_i["name"],
                    feature_type=feature_i["featureType"],
                )
                for feature_i in model.independentFeatures
            ],
            visibility=ModelVisibility(visibility),
            actual_model=actual_model,
            description=description,
        )

        response = create_model.sync_detailed(client=auth_client, body=body_model)
        if response.status_code < 300:
            model_url = response.headers.get("Location")
            model_id = model_url.split("/")[-1]

            self.log.info(
                "Model has been successfully uploaded. The url of the model is %s",
                self.app_url + "/dashboard/models/" + model_id,
            )
        else:
            # error = response.headers.get("error")
            # error_description = response.headers.get("error_description")
            self.log.error("Error code: " + str(response.status_code.value))


def deploy_Torch_Graph_model(self, deployment_json):
    auth_client = AuthenticatedClient(base_url=self.api_url, token=self.api_key)
    body_model = Model(
        name=deployment_json["actualModel"],
        type=deployment_json["model_type"],
        jaqpotpy_version="",
        libraries="",
        dependent_features=deployment_json["dependentFeatures"],
        independent_features=deployment_json["independentFeatures"],
        visibility=deployment_json["visibility"],
        actual_model=deployment_json["actualModel"],
        description=deployment_json["description"],
    )

    response = create_model.sync_detailed(client=auth_client, body=body_model)
    if response.status_code < 300:
        self.log.info(
            "Model has been successfully uploaded. The url of the model is "
            + response.headers.get("Location")
        )
    else:
        # error = response.headers.get("error")
        # error_description = response.headers.get("error_description")
        self.log.error("Error code: " + str(response.status_code.value))

    def deploy_XGBoost(self, model, X, y, title, description, algorithm, doa=None):
        """Deploys XGBoost model to Jaqpot.

        Extended description of function.

        Parameters
        ----------
        model : XGBoost trained model
            model is a trained model that occurs from the XGBoost family of algorithms
        X : pandas dataframe
            The dataframe that is used to train the model (X variables).
        y : pandas dataframe
            The dataframe that is used to train the model (y variables).
        title: String
            The title of the model
        description: String
            The description of the model
        algorithm: String
            The algorithm that the model implements
        doa: pandas dataframe
            The dataset used to create the domain of applicability of the model

        Returns
        -------
        string
            The id of the model that uploaded

        """
        if isinstance(X, pd.DataFrame) is False:
            raise Exception(
                "Function deploy_glm supports pandas dataframe or series. X is not one"
            )
        if isinstance(y, pd.DataFrame) is False and isinstance(y, pd.Series) is False:
            raise Exception(
                "Function deploy_glm supports pandas dataframe or series. Y is not one"
            )
        # coef_flatten = model.coef_.flatten()
        # intercept_flatten = model.intercept_.flatten()
        # coef_df = pd.DataFrame(list(zip(list(X), coef_flatten)), columns=['Features', 'coeff'])
        # intercept = pd.DataFrame([('(intercept)', intercept_flatten[0])], columns=['Features', 'coeff'])
        # coefs_all = intercept.append(coef_df, ignore_index=True)
        # coefs = {}
        additionalInfo = {}
        # for key in coefs_all.values:
        #     coefs[key[0]] = key[1]
        # additionalInfo['coefficients'] = coefs
        additionalInfo["inputSeries"] = list(X)
        pretrained = help.create_pretrain_req(
            model,
            X,
            y,
            title,
            description,
            algorithm,
            "XGBoost model",
            "XGBoost-model",
            additionalInfo,
        )
        # j = json.dumps(pretrained)
        j = json.dumps(pretrained, cls=JaqpotSerializer)
        if doa is None:
            response = models_api.post_pretrained_model(
                self.base_url, self.api_key, j, self.log
            )
            if response.status_code < 300:
                resp = response.json()
                self.log.info(
                    "Model with id: "
                    + resp["modelId"]
                    + " created. Please visit the application to proceed"
                )
                return resp["modelId"]
            else:
                resp = response.json()
                self.log.error("Some error occured: " + resp["message"])
                return
        else:
            response = models_api.post_pretrained_model(
                self.base_url, self.api_key, j, self.log
            )
            if response.status_code < 300:
                resp = response.json()
                self.log.info(
                    "Model with id: "
                    + resp["modelId"]
                    + " created. Storing Domain of applicability"
                )
                modid = resp["modelId"]
            else:
                resp = response.json()
                self.log.error("Some error occured: " + resp["message"])
                return
            # loop = asyncio.get_event_loop()
            # a = loop.create_task(jha.calculate_a(X))
            # b = loop.create_task(jha.calculate_doa_matrix(X))

            a = jha.calculate_a(X)
            b = jha.calculate_doa_matrix(X)
            # all_groups = asyncio.gather(a, b)
            # results = loop.run_until_complete(all_groups)
            doa = help.create_doa(inv_m=b.values.tolist(), a=a, modelid=resp["modelId"])
            j = json.dumps(doa, cls=JaqpotSerializer)
            resp = doa_api.post_models_doa(self.base_url, self.api_key, j, self.log)
            if resp == 201:
                self.log.info(
                    "Stored Domain of applicability. Visit the application to proceed"
                )
                return modid

    def get_model_by_id(self, model):
        """Retrieves user's model by ID.

        Parameters
        ----------
        model : str
            The model's ID.


        Returns
        -------
        Object
            The particular model.

        """
        return models_api.get_model(self.base_url, self.api_key, model, self.log)

    def get_feature_by_id(self, feature):
        """Retrieves a Jaqpot feature.

        Parameters
        ----------
        feature : str
            The feature's ID.


        Returns
        -------
        Object
            The particular feature.

        """
        return featapi.get_feature(self.base_url, self.api_key, feature)

    def get_my_models(self, minimum, maximum):
        """Retrieves user's models.

        Parameters
        ----------
        minimum : int
            The index of the first model.
        maximum : int
            The index of the last model.

        Returns
        -------
        Object
            The models of the user.

        """
        return models_api.get_my_models(
            self.base_url, self.api_key, minimum, maximum, self.log
        )

    def get_orgs_models(self, organization, minimum, maximum):
        """Retrieves organization's models.

        Parameters
        ----------
        organization: str
            The organization's ID.
        minimum : int
            The index of the first model.
        maximum : int
            The index of the last model.

        Returns
        -------
        Object
            The models of the organization.

        """
        return models_api.get_orgs_models(
            self.base_url, self.api_key, organization, minimum, maximum, self.log
        )

    def get_models_by_tag(self, tag, minimum, maximum):
        """Retrieves models with a particular tag.

        Parameters
        ----------
        tag: str
            The model's tag.
        minimum : int
            The index of the first model.
        maximum : int
            The index of the last model.

        Returns
        -------
        Object
            The models of the organization.

        """
        return models_api.get_models_by_tag(
            self.base_url, self.api_key, tag, minimum, maximum, self.log
        )

    def get_models_by_tag_and_org(self, organization, tag, minimum, maximum):
        """Retrieves models of an organization with a particular tag.

        Parameters
        ----------
        organization: str:
            The organization's ID.
        tag: str
            The model's tag.
        minimum : int
            The index of the first model.
        maximum : int
            The index of the last model.

        Returns
        -------
        Object
            The models of the organization.

        """
        return models_api.get_models_by_tag(
            self.base_url, self.api_key, organization, tag, minimum, maximum, self.log
        )

    def get_dataset(self, dataset):
        """Retrieves a dataset.

        Parameters
        ----------
        dataset: str
            The dataset's ID.

        Returns
        -------
        pd.DataFrame()
            DataFrame of the dataset.

        Note
        ----
        Mazimum rows retrieved: 1000.

        """
        dataRetrieved = data_api.get_dataset(
            self.base_url, self.api_key, dataset, self.log
        )

        feats = ["" for i in range(len(dataRetrieved["features"]))]

        for item in dataRetrieved["features"]:
            feats[int(item["key"])] = item["name"]

        indices = [item["entryId"]["name"] for item in dataRetrieved["dataEntry"]]

        data = [
            [item["values"][str(feats.index(col))] for col in feats]
            for item in dataRetrieved["dataEntry"]
        ]

        return pd.DataFrame(data, index=indices, columns=feats)

    def get_doa(self, model):
        """Retrieves model's domain of applicability.

        Parameters
        ----------
        model: str
            The model's ID.

        Returns
        -------
        Object
            The model's domain of applicability.

        """
        return doa_api.get_models_doa(self.base_url, self.api_key, model, self.log)



from rdkit import DataStructs
from rdkit import Chem
from rdkit.Chem import QED
from rdkit.Chem import Crippen
import pickle
import gzip
import numpy as np
import math

import pkg_resources
from typing import Iterable, Any

NP_file = pkg_resources.resource_filename(
    "jaqpotpy.models.generative.data", "NP_score.pkl.gz"
)
SA_file = pkg_resources.resource_filename(
    "jaqpotpy.models.generative.data", "SA_score.pkl.gz"
)

NP_model = pickle.load(gzip.open(NP_file))
SA_model = {
    i[j]: float(i[0]) for i in pickle.load(gzip.open(SA_file)) for j in range(1, len(i))
}


def validate_molecule(mol):
    try:
        x = Chem.MolToSmiles(mol)
    except Exception:
        x = None
    return x


def novel_score(mols: Iterable[Any], smiles: Iterable[str]):
    valid_mols = []
    valid_smiles = []
    for mol in mols:
        smiles_v = validate_molecule(mol)
        if smiles_v:
            valid_smiles.append(smiles_v)
            valid_mols.append(Chem.MolFromSmiles(smiles_v))
        else:
            valid_mols.append(None)
            valid_smiles.append(None)
    ar = np.full((1, len(valid_mols)), True, dtype=bool)
    for index, v_smile in enumerate(valid_smiles):
        if v_smile not in smiles:
            ar[0][index] = v_smile
    return ar[0]


def novel_mean_score(mols, data):
    valid = novel_score(mols, data)
    return valid.mean()


def valid_mean(mols):
    valids = []
    for m in mols:
        try:
            valids.append(1)
        except Exception:
            valids.append(0)
    return np.array(valids).mean()


def valids(mols):
    # return x is not None and Chem.MolToSmiles(x) != ''
    valids = []
    try:
        for m in mols:
            try:
                # smiles = Chem.MolToSmiles(m)
                valids.append(m)
            except Exception:
                valids.append(None)
    except TypeError:
        try:
            valids.append(mols)
        except Exception:
            valids.append(None)
    return valids


def compute_SAS(mol):
    if mol is None:
        return 0
    fp = Chem.rdMolDescriptors.GetMorganFingerprint(mol, 2)
    fps = fp.GetNonzeroElements()
    score1 = 0.0
    nf = 0
    # for bitId, v in fps.items():
    for bitId, v in fps.items():
        nf += v
        sfp = bitId
        score1 += SA_model.get(sfp, -4) * v
    try:
        score1 /= nf
    except ZeroDivisionError:
        return 0

    # features score
    nAtoms = mol.GetNumAtoms()
    nChiralCenters = len(Chem.FindMolChiralCenters(mol, includeUnassigned=True))
    ri = mol.GetRingInfo()
    nSpiro = Chem.rdMolDescriptors.CalcNumSpiroAtoms(mol)
    nBridgeheads = Chem.rdMolDescriptors.CalcNumBridgeheadAtoms(mol)
    nMacrocycles = 0
    for x in ri.AtomRings():
        if len(x) > 8:
            nMacrocycles += 1

    sizePenalty = nAtoms ** 1.005 - nAtoms
    stereoPenalty = math.log10(nChiralCenters + 1)
    spiroPenalty = math.log10(nSpiro + 1)
    bridgePenalty = math.log10(nBridgeheads + 1)
    macrocyclePenalty = 0.0

    # ---------------------------------------
    # This differs from the paper, which defines:
    #  macrocyclePenalty = math.log10(nMacrocycles+1)
    # This form generates better results when 2 or more macrocycles are present
    if nMacrocycles > 0:
        macrocyclePenalty = math.log10(2)

    score2 = (
            0.0
            - sizePenalty
            - stereoPenalty
            - spiroPenalty
            - bridgePenalty
            - macrocyclePenalty
    )

    # correction for the fingerprint density
    # not in the original publication, added in version 1.1
    # to make highly symmetrical molecules easier to synthetise
    score3 = 0.0
    if nAtoms > len(fps):
        score3 = math.log(float(nAtoms) / len(fps)) * 0.5

    sascore = score1 + score2 + score3

    # need to transform "raw" value into scale between 1 and 10
    min = -4.0
    max = 2.5
    sascore = 11.0 - (sascore - min + 1) / (max - min) * 9.0
    # smooth the 10-end
    if sascore > 8.0:
        sascore = 8.0 + math.log(sascore + 1.0 - 9.0)
    if sascore > 10.0:
        sascore = 10.0
    elif sascore < 1.0:
        sascore = 1.0

    return sascore


def _remap(x, x_min, x_max):
    return (x - x_min) / (x_max - x_min)


def synthetic_accessibility_score_scores(mols, norm=False):
    scores = [compute_SAS(mol) if mol is not None else None for mol in mols]
    scores = np.array(list(map(lambda x: 0 if x is None else x, scores)))
    scores = np.clip(_remap(scores, 5, 1.5), 0.0, 1.0) if norm else scores
    return scores


def natural_products_score(mols, norm=False):
    # calculating the score
    scores = [
        sum(
            NP_model.get(bit, 0)
            for bit in Chem.rdMolDescriptors.GetMorganFingerprint(
                mol, 2
            ).GetNonzeroElements()
        )
        / float(mol.GetNumAtoms())
        if mol is not None
        else None
        for mol in mols
    ]

    # preventing score explosion for exotic molecules
    scores = list(
        map(
            lambda score: score
            if score is None
            else (
                4 + math.log10(score - 4 + 1)
                if score > 4
                else (-4 - math.log10(-4 - score + 1) if score < -4 else score)
            ),
            scores,
        )
    )

    scores = np.array(list(map(lambda x: -4 if x is None else x, scores)))
    scores = np.clip(_remap(scores, -3, 1), 0.0, 1.0) if norm else scores

    return scores


def _avoid_sanitization_error(op):
    try:
        return op()
    except ValueError:
        return None


def quantitative_estimation_druglikeness_scores(mols, norm=True):
    return np.array(
        list(
            map(
                lambda x: 0 if x is None else x,
                [
                    _avoid_sanitization_error(lambda: QED.qed(mol))
                    if mol is not None
                    else None
                    for mol in mols
                ],
            )
        )
    )


def water_octanol_partition_coefficient_scores(mols, norm=False):
    scores = [
        _avoid_sanitization_error(lambda: Crippen.MolLogP(mol))
        if mol is not None
        else None
        for mol in mols
    ]
    scores = np.array(list(map(lambda x: -3 if x is None else x, scores)))
    scores = (
        np.clip(_remap(scores, -2.12178879609, 6.0429063424), 0.0, 1.0)
        if norm
        else scores
    )
    return scores


def _compute_diversity(mol, fps):
    ref_fps = Chem.rdMolDescriptors.GetMorganFingerprintAsBitVect(mol, 4, nBits=2048)
    dist = DataStructs.BulkTanimotoSimilarity(ref_fps, fps, returnDistance=True)
    score = np.mean(dist)
    return score


def diversity_scores(mols, data):
    rand_mols = np.random.choice(data, 100)
    fps = [
        Chem.rdMolDescriptors.GetMorganFingerprintAsBitVect(mol, 4, nBits=2048)
        for mol in rand_mols
    ]

    scores = np.array(
        list(map(lambda x: _compute_diversity(x, fps) if x is not None else 0, mols))
    )
    scores = np.clip(_remap(scores, 0.9, 0.945), 0.0, 1.0)
    return scores


def unique_total_score(mols):
    v = valids(mols)
    s = set()
    for mol in v:
        try:
            smile = Chem.MolToSmiles(mol)
            s.add(smile)
        except Exception:
            continue
    # s = set(map(lambda x: Chem.MolToSmiles(x), v))
    return 0 if len(v) == 0 else len(s) / len(v)


def _constant_bump(x, x_low, x_high, decay=0.025):
    return np.select(
        condlist=[x <= x_low, x >= x_high],
        choicelist=[
            np.exp(-((x - x_low) ** 2) / decay),
            np.exp(-((x - x_high) ** 2) / decay),
        ],
        default=np.ones_like(x),
    )


def novel_scores(mols, data):
    smiles = []
    for mol in mols:
        try:
            sm = Chem.MolToSmiles(mol)
            smiles.append(sm)
        except Exception:
            smiles.append(None)
    novel_l = []
    for smile in smiles:
        if smile not in data and smile is not None:
            novel_l.append(True)
        else:
            novel_l.append(False)
    novels = np.array(novel_l)
    return novels


def drugcandidate_scores(mols, data):
    scores = (
                     _constant_bump(
                         water_octanol_partition_coefficient_scores(mols, norm=True), 0.210, 0.945
                     )
                     + synthetic_accessibility_score_scores(mols, norm=True)
                     + novel_scores(mols, data)
                     + (1 - novel_scores(mols, data)) * 0.3
             ) / 4

    return scores


def valid_scores(mols):
    t = map(valids, mols)
    li = list(t)
    trues = []
    for mo in li:
        if mo[0] is not None:
            trues.append(True)
        else:
            trues.append(False)
    score = np.array(trues, dtype=np.float32)
    return score


class MolecularMetrics(object):
    @staticmethod
    def _avoid_sanitization_error(op):
        try:
            return op()
        except ValueError:
            return None

    @staticmethod
    def remap(x, x_min, x_max):
        return (x - x_min) / (x_max - x_min)

    @staticmethod
    def valid_lambda(x):
        try:
            smiles = Chem.MolToSmiles(x)
        except Exception:
            x = None
        return x is not None and smiles != ""

    @staticmethod
    def valid_lambda_special(x):
        s = Chem.MolToSmiles(x) if x is not None else ""
        return x is not None and "*" not in s and "." not in s and s != ""

    @staticmethod
    def valid_scores(mols):
        return np.array(
            list(map(MolecularMetrics.valid_lambda, mols)), dtype=np.float32
        )

    @staticmethod
    def valid_filter(mols):
        return list(filter(MolecularMetrics.valid_lambda, mols))

    @staticmethod
    def valid_total_score(mols):
        return np.array(
            list(map(MolecularMetrics.valid_lambda, mols)), dtype=np.float32
        ).mean()

    @staticmethod
    def novel_scores(mols, smiles):
        return np.array(
            list(
                map(
                    lambda x: MolecularMetrics.valid_lambda(x)
                              and Chem.MolToSmiles(x) not in smiles,
                    mols,
                )
            )
        )

    @staticmethod
    def novel_filter(mols, smiles):
        return list(
            filter(
                lambda x: MolecularMetrics.valid_lambda(x)
                          and Chem.MolToSmiles(x) not in smiles,
                mols,
            )
        )

    @staticmethod
    def novel_total_score(mols, data):
        return MolecularMetrics.novel_scores(
            MolecularMetrics.valid_filter(mols), data
        ).mean()

    @staticmethod
    def unique_scores(mols):
        smiles = list(
            map(
                lambda x: Chem.MolToSmiles(x)
                if MolecularMetrics.valid_lambda(x)
                else "",
                mols,
            )
        )
        return np.clip(
            0.75
            + np.array(
                list(map(lambda x: 1 / smiles.count(x) if x != "" else 0, smiles)),
                dtype=np.float32,
            ),
            0,
            1,
        )

    @staticmethod
    def unique_total_score(mols):
        v = MolecularMetrics.valid_filter(mols)
        s = set(map(lambda x: Chem.MolToSmiles(x), v))
        return 0 if len(v) == 0 else len(s) / len(v)

    @staticmethod
    def natural_product_scores(mols, norm=False):
        # calculating the score
        scores = [
            sum(
                NP_model.get(bit, 0)
                for bit in Chem.rdMolDescriptors.GetMorganFingerprint(
                    mol, 2
                ).GetNonzeroElements()
            )
            / float(mol.GetNumAtoms())
            if mol is not None
            else None
            for mol in mols
        ]

        # preventing score explosion for exotic molecules
        scores = list(
            map(
                lambda score: score
                if score is None
                else (
                    4 + math.log10(score - 4 + 1)
                    if score > 4
                    else (-4 - math.log10(-4 - score + 1) if score < -4 else score)
                ),
                scores,
            )
        )

        scores = np.array(list(map(lambda x: -4 if x is None else x, scores)))
        scores = (
            np.clip(MolecularMetrics.remap(scores, -3, 1), 0.0, 1.0) if norm else scores
        )

        return scores

    @staticmethod
    def quantitative_estimation_druglikeness_scores(mols, norm=False):
        return np.array(
            list(
                map(
                    lambda x: 0 if x is None else x,
                    [
                        MolecularMetrics._avoid_sanitization_error(lambda: QED.qed(mol))
                        if mol is not None
                        else None
                        for mol in mols
                    ],
                )
            )
        )

    @staticmethod
    def water_octanol_partition_coefficient_scores(mols, norm=False):
        scores = [
            MolecularMetrics._avoid_sanitization_error(lambda: Crippen.MolLogP(mol))
            if mol is not None
            else None
            for mol in mols
        ]
        scores = np.array(list(map(lambda x: -3 if x is None else x, scores)))
        scores = (
            np.clip(
                MolecularMetrics.remap(scores, -2.12178879609, 6.0429063424), 0.0, 1.0
            )
            if norm
            else scores
        )

        return scores

    # @staticmethod
    # def diversity_scores(mols, data):
    #     rand_mols = np.random.choice(data, 100)
    #     fps = [Chem.rdMolDescriptors.GetMorganFingerprintAsBitVect(mol, 4, nBits=2048) for mol in rand_mols]
    #
    #     scores = np.array(
    #         list(map(lambda x: MolecularMetrics.__compute_diversity(x, fps) if x is not None else 0, mols)))
    #     scores = np.clip(MolecularMetrics.remap(scores, 0.9, 0.945), 0.0, 1.0)
    #
    #     return scores

    @staticmethod
    def synthetic_accessibility_score_scores(mols, norm=False):
        scores = [
            MolecularMetrics._compute_SAS(mol) if mol is not None else None
            for mol in mols
        ]
        scores = np.array(list(map(lambda x: 10 if x is None else x, scores)))
        scores = (
            np.clip(MolecularMetrics.remap(scores, 5, 1.5), 0.0, 1.0)
            if norm
            else scores
        )

        return scores

    @staticmethod
    def __compute_diversity(mol, fps):
        ref_fps = Chem.rdMolDescriptors.GetMorganFingerprintAsBitVect(
            mol, 4, nBits=2048
        )
        dist = DataStructs.BulkTanimotoSimilarity(ref_fps, fps, returnDistance=True)
        score = np.mean(dist)
        return score

    @staticmethod
    def drugcandidate_scores(mols, data):
        scores = (
                         MolecularMetrics.constant_bump(
                             MolecularMetrics.water_octanol_partition_coefficient_scores(
                                 mols, norm=True
                             ),
                             0.210,
                             0.945,
                         )
                         # + MolecularMetrics.synthetic_accessibility_score_scores(mols,
                         #                                                          norm=True)
                         + MolecularMetrics.novel_scores(mols, data)
                         + (1 - MolecularMetrics.novel_scores(mols, data)) * 0.3
                 ) / 4

        return scores

    @staticmethod
    def constant_bump(x, x_low, x_high, decay=0.025):
        return np.select(
            condlist=[x <= x_low, x >= x_high],
            choicelist=[
                np.exp(-((x - x_low) ** 2) / decay),
                np.exp(-((x - x_high) ** 2) / decay),
            ],
            default=np.ones_like(x),
        )



from jaqpotpy.descriptors.base_classes import MaterialFeaturizer
from pymatgen.core.structure import Structure
from typing import Union, Tuple
import numpy as np
import scipy.constants as const
import pandas as pd

ANG_TO_BOHR = const.value("Angstrom star") / const.value("Bohr radius")


class SineCoulombMatrix(MaterialFeaturizer):
    """Create a fixed vector of length 118, containing rwa, fractional
    element compostitions in a compound.

    References:
    ----------
    Jha, D., Ward, L., Paul, A. et al. Sci Rep 8, 17593 (2018).
       https://doi.org/10.1038/s41598-018-35934-y

    Examples:
    --------
    >>> import jaqpotpy as jt
    >>> from pymatgen.core.lattice import Lattice
    >>> from pymatgen.core.structure import Structure
    >>> lattice = Lattice.cubic(4.2)
    >>> struct = Structure(lattice, ["Cs", "Cl"], [[0, 0, 0], [0.5, 0.5, 0.5]])
    >>> featurizer = jt.descriptors.material.SineCoulombMatrix()
    >>> features = featurizer.featurize(struct)
    >>> type(features[0])
    <class 'numpy.ndarray'>
    >>> features[0].shape
    (118,)

    Note:
    ----
    This class requires matminer and Pymatgen to be installed.
    `NaN` feature values are automatically converted to 0 by this featurizer.

    """

    def __init__(
        self, max_atoms: int = 100, diag_elements: bool = True, flatten: bool = False
    ):
        """Create a Sine Coulomb Matrix for a pymatgen.core.Structure

        Parameters
        ----------
        max_atoms: int (default 100)
          Maximum number of atoms for any crystal in the dataset. Used to
          pad the Coulomb matrix.
        diag_elems (bool): flag indication whether (True, default) to use
          the original definition of the diagonal elements; if set to False,
          the diagonal elements are set to 0
        flatten: bool (default True)
          Return flattened vector of matrix eigenvalues.

        References
        ----------
          Faber et al. "Crystal Structure Representations for Machine
             Learning Models of Formation Energies", Inter. J. Quantum Chem.
             115, 16, 2015. https://arxiv.org/abs/1503.07406

        """
        self.diag_elements = diag_elements
        self.flatten = flatten
        self.max_atoms = max_atoms
        return

    @property
    def __name__(self):
        return "SineCoulombMatrix"

    def __getitem__(self):
        return self

    def _featurize(self, datapoint: Union[str, Structure], **kwargs) -> np.ndarray:
        """Calculate sine coulomb matrix for a structure.

        Parameters
        ----------
        datapoint: str (path of a file) or pymatgen.core.Structure
          Structure or path of a file to parse into a Structure

        Returns
        -------
        feats: np.ndarray
          Vector containing the sine coulomb matrix of a structure.

        """
        if isinstance(datapoint, str):
            if datapoint.split(".")[-1] == "extxyz":
                from ase.io import read

                xyz = read(datapoint)
                lat = np.array(xyz.get_cell())
                sym = xyz.get_chemical_symbols()
                pos = xyz.arrays["positions"]
                struct = Structure(lat, sym, pos)
            else:
                struct = Structure.from_file(datapoint)
        else:
            struct = datapoint

        sites = struct.sites
        atomic_numbers = np.array([site.specie.Z for site in sites])
        sin_mat = np.zeros((len(sites), len(sites)))
        coords = np.array([site.frac_coords for site in sites])
        lattice = struct.lattice.matrix

        for i in range(len(sin_mat)):
            for j in range(len(sin_mat)):
                if i == j:
                    if self.diag_elements:
                        sin_mat[i][i] = 0.5 * atomic_numbers[i] ** 2.4
                elif i < j:
                    vec = coords[i] - coords[j]
                    coord_vec = np.sin(np.pi * vec) ** 2
                    trig_dist = (
                        np.linalg.norm((np.asmatrix(coord_vec) * lattice).A1)
                        * ANG_TO_BOHR
                    )
                    sin_mat[i][j] = atomic_numbers[i] * atomic_numbers[j] / trig_dist
                else:
                    sin_mat[i][j] = sin_mat[j][i]

        if self.flatten:
            eigs, _ = np.linalg.eig([sin_mat])
            zeros = np.zeros(self.max_atoms)
            zeros[: len(eigs[0])] = eigs[0]
            features = zeros
        else:
            features = pad_array([sin_mat], self.max_atoms)

        features = np.asarray(features)
        return features

    def _featurize_dataframe(
        self, datapoint: Union[str, Structure], **kwargs
    ) -> pd.DataFrame:
        """Calculate sine coulomb matrix for a structure.

        Parameters
        ----------
        datapoint: str (path of a file) or pymatgen.core.Structure
          Structure or path of a file to parse into a Structure

        Returns
        -------
        feats: pd.DataFrame
          Vector containing the sine coulomb matrix of a structure.

        """
        self.flatten = True
        data = self._featurize(datapoint)
        columns = ["Feat_{}".format(str(i)) for i in range(self.max_atoms)]
        return pd.DataFrame([data], columns=columns, index=[0])


def pad_array(
    x: list, shape: Union[Tuple, int], fill: float = 0.0, both: bool = False
) -> np.ndarray:
    """Pad an array with a fill value.

    Parameters
    ----------
    x: np.ndarray
      A numpy array.
    shape: Tuple or int
      Desired shape. If int, all dimensions are padded to that size.
    fill: float, optional (default 0.0)
      The padded value.
    both: bool, optional (default False)
      If True, split the padding on both sides of each axis. If False,
      padding is applied to the end of each axis.

    Returns
    -------
    np.ndarray
      A padded numpy array

    """
    x = np.asarray(x)
    if not isinstance(shape, tuple):
        shape = tuple(shape for _ in range(x.ndim))
    pad = []
    for i in range(x.ndim):
        diff = shape[i] - x.shape[i]
        assert diff >= 0
        if both:
            a, b = divmod(diff, 2)
            b += a
            pad.append((a, b))
        else:
            pad.append((0, diff))
    pad = tuple(pad)  # type: ignore
    x = np.pad(x, pad, mode="constant", constant_values=fill)
    return x



from typing import Callable, List, Optional, Any
import numpy as np

# from deepchem.utils.typing import RDKitMol
from jaqpotpy.descriptors.base_classes import MolecularFeaturizer

RDKitMol = Any
RDKitAtom = Any
RDKitBond = Any


# warnings.filterwarnings("ignore", category=DeprecationWarning)


class MordredDescriptors(MolecularFeaturizer):
    """Mordred descriptors.
    This class computes a list of chemical descriptors using Mordred.
    Please see the details about all descriptors from [1]_, [2]_.

    Attributes:
    ----------
    descriptors: List[str]
      List of Mordred descriptor names used in this class.

    References:
    ----------
    .. [1] Moriwaki, Hirotomo, et al. "Mordred: a molecular descriptor calculator."
       Journal of cheminformatics 10.1 (2018): 4.
    .. [2] http://mordred-descriptor.github.io/documentation/master/descriptors.html

    Note:
    ----
    This class requires Mordred to be installed.

    Examples:
    --------
    >>> import jaqpotpy as jt
    >>> smiles = ['CC(=O)OC1=CC=CC=C1C(=O)O']
    >>> featurizer = jt.descriptors.MordredDescriptors(ignore_3D=True)
    >>> features = featurizer.featurize(smiles)
    >>> type(features[0])
    <class 'numpy.ndarray'>
    >>> features[0].shape
    (1613,)

    """

    @property
    def __name__(self):
        return "MordredDescriptors"

    def __init__(self, ignore_3D: bool = True):
        """Parameters
        ----------
        ignore_3D: bool, optional (default True)
          Whether to use 3D information or not.

        """
        self.ignore_3D = ignore_3D
        self.calc: Optional[Callable] = None
        self.descriptors: Optional[List] = None

    def __getitem__(self):
        return self

    def _featurize(
            self, datapoint: RDKitMol, convert_nan: bool = True, **kwargs
    ) -> np.ndarray:
        """Calculate Mordred descriptors.

        Parameters
        ----------
        datapoint: rdkit.Chem.rdchem.Mol
          RDKit Mol object

        Returns
        -------
        np.ndarray
          1D array of Mordred descriptors for `mol`.
          If ignore_3D is True, the length is 1613.
          If ignore_3D is False, the length is 1826.

        """
        if "mol" in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )
        if self.calc is None:
            try:
                from mordred import Calculator, descriptors, is_missing

                self.is_missing = is_missing
                self.calc = Calculator(descriptors, ignore_3D=self.ignore_3D)
                self.descriptors = list(descriptors.__all__)
            except ModuleNotFoundError:
                raise ImportError("This class requires Mordred to be installed.")

        feature = self.calc(datapoint)
        # convert errors to zero
        if convert_nan:
            feature = [
                -1000.0 if self.is_missing(val) or isinstance(val, str) else val
                for val in feature
            ]
        return np.asarray(feature)

    def _get_column_names(self, **kwargs) -> list:
        """Return the column names"""
        if "mol" in kwargs:
            # datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )
        if self.calc is None:
            try:
                from mordred import Calculator, descriptors, is_missing

                self.is_missing = is_missing
                self.calc = Calculator(descriptors, ignore_3D=self.ignore_3D)
                self.descriptors = list(descriptors.__all__)
            except ModuleNotFoundError:
                raise ImportError("This class requires Mordred to be installed.")
        from rdkit import Chem

        smiles = "CC(=O)OC1=CC=CC=C1C(=O)O"
        mol = Chem.MolFromSmiles(smiles)
        feature = self.calc(mol).fill_missing(0).asdict()
        names = [key for key in feature.keys()]

        return names

    def _featurize_dataframe(
            self, datapoint: RDKitMol, convert_nan: bool = True, **kwargs
    ) -> np.ndarray:
        """Calculate Mordred descriptors.

        Parameters
        ----------
        datapoint: rdkit.Chem.rdchem.Mol
        RDKit Mol object

        Returns
        -------
        np.ndarray
        1D array of Mordred descriptors for `mol`.
        If ignore_3D is True, the length is 1613.
        If ignore_3D is False, the length is 1826.

        """
        if "mol" in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )
        if self.calc is None:
            try:
                from mordred import Calculator, descriptors, is_missing

                self.is_missing = is_missing
                self.calc = Calculator(descriptors, ignore_3D=self.ignore_3D)
                self.descriptors = list(descriptors.__all__)
            except ModuleNotFoundError:
                raise ImportError("This class requires Mordred to be installed.")
        feature = self.calc(datapoint)
        # convert errors to zero
        if convert_nan:
            feature = [
                -1000.0 if self.is_missing(val) or isinstance(val, str) else val
                for val in feature
            ]
        return np.asarray(feature)



"""Tests for Docking"""

import os
import platform
import unittest

# import pytest
import jaqpotpy as jp

IS_WINDOWS = platform.system() == "Windows"


class TestDockingUtils(unittest.TestCase):
    """Does sanity checks on pose generation."""

    def setUp(self):
        current_dir = os.path.dirname(os.path.realpath(__file__))
        # self.protein_file = os.path.join(current_dir, "1jld_protein.pdb")
        # self.protein_file = os.path.join(current_dir, "7zb6.pdb")
        # self.ligand_file = os.path.join(current_dir, "mulno.sdf")
        # self.ligand_file = os.path.join(current_dir, "1jld_ligand.sdf")

        self.protein_file = os.path.join(current_dir, "1a9m_pocket.pdb")
        self.ligand_file = os.path.join(current_dir, "1a9m_ligand.sdf")

    # @pytest.mark.slow
    @unittest.skip("Docking has not been integrated in the latest jaqpotpy version")
    def test_docker_init(self):
        """Test that Docker can be initialized."""
        vpg = jp.docking.VinaPoseGenerator()
        jp.docking.Docker(vpg)

    @unittest.skip("test only for conda env")
    def test_docking_utils(self):
        from jaqpotpy.docking.utils import create_hydrated_pdbqt_pdb

        protein_pdbqt = create_hydrated_pdbqt_pdb(self.protein_file, "./tmp")
        print(protein_pdbqt)

    @unittest.skipIf(IS_WINDOWS, "vina is not supported in windows")
    @unittest.skip("Docking has not been integrated in the latest jaqpotpy version")
    # @pytest.mark.slow
    def test_docker_dock(self):
        """Test that Docker can dock."""
        # We provide no scoring model so the docker won't score
        vpg = jp.docking.VinaPoseGenerator()
        docker = jp.docking.Docker(vpg)
        docked_outputs = docker.dock(
            (self.protein_file, self.ligand_file),
            exhaustiveness=10,
            num_modes=10,
            out_dir="./tmp",
        )

        print(docked_outputs)
        # Check only one output since num_modes==1
        assert len(list(docked_outputs)) == 10



"""Author: Ioannis Pitoskas (jpitoskas@gmail.com)"""


class Feature:
    @staticmethod
    def get_feature_names_and_possible_values_from_column_names(input_list):
        structured_data = {}

        for column_name in input_list:
            if "." in column_name:
                feature_name, categorical_value = column_name.split(".")
                if feature_name not in structured_data:
                    structured_data[feature_name] = []
                structured_data[feature_name].append(categorical_value)
            else:
                structured_data[column_name] = []

        feature_names_and_possible_values = [
            {"name": feature_name, "possibleValues": possible_values}
            for feature_name, possible_values in structured_data.items()
        ]

        return feature_names_and_possible_values

    def __init__(
        self,
        name,
        featureDependency,
        possibleValues,
        featureType,
        meta=None,
        description=None,
    ):
        if not isinstance(name, str):
            raise ValueError("'name' should be of type str")

        if featureDependency not in ["DEPENDENT", "INDEPENDENT"]:
            raise ValueError(
                f"possible values for 'featureDependency' are {['DEPENDENT', 'INDEPENDENT']}"
            )

        if possibleValues is not None and not isinstance(possibleValues, list):
            raise ValueError("'possibleValues' should be of type list")

        if possibleValues is not None:
            for possibleValue in possibleValues:
                if not isinstance(possibleValue, str):
                    raise ValueError("items in 'possibleValues' should be of type str")

        if meta is not None and not isinstance(meta, dict):
            raise ValueError("'meta' should be of type dict")

        featureType_allowed_values = [
            "CATEGORICAL",
            "SMILES",
            "INTEGER",
            "FLOAT",
            "TEXT",
        ]
        if featureType is not None and featureType not in featureType_allowed_values:
            raise ValueError(
                f"possible values for 'featureType' are {featureType_allowed_values}"
            )

        if description is not None and not isinstance(description, str):
            raise ValueError("'description' should be of type str")

        self.name = name
        self.featureDependency = featureDependency

        self.possibleValues = possibleValues if possibleValues is not None else []

        if self.possibleValues == [] and featureType is None:
            self.featureType = "FLOAT"
        if self.possibleValues != [] and featureType is None:
            self.featureType = "CATEGORICAL"
        if self.possibleValues != [] and featureType in ["FLOAT", "SMILES"]:
            raise ValueError(
                f"Feature of featureType '{featureType}' cannot have a finite set of possible values"
            )
        if self.possibleValues == [] and featureType in ["CATEGORICAL"]:
            raise ValueError(
                f"Feature of featureType '{featureType}' must have a non-empty set of possible values"
            )

        self.featureType = featureType

        self.meta = meta if meta is not None else {}
        self.description = description if description is not None else ""

    def to_json(self):
        feature_dict = {
            "meta": self.meta,
            "name": self.name,
            "description": self.description,
            "featureType": self.featureType,
            "featureDependency": self.featureDependency,
            "possibleValues": self.possibleValues,
        }
        return feature_dict

    def __str__(self):
        return self.to_json()

    def __repr__(self):
        return f"{self.__class__.__name__}({self.to_json()})"



from jaqpotpy.descriptors.base_classes import MaterialFeaturizer
from pymatgen.core.composition import Composition
from pymatgen.util.typing import CompositionLike
from typing import Any
from jaqpotpy.helpers.statistics import PropertyStats
import numpy as np
import pandas as pd


class ElementPropertyFingerprint(MaterialFeaturizer):
    """Fingerprint of elemental properties from composition.
    Based on the data source chosen, returns properties and statistics
    (min, max, range, mean, standard deviation, mode) for a compound
    based on elemental stoichiometry. E.g., the average electronegativity
    of atoms in a crystal structure. The chemical fingerprint is a
    vector of these statistics.
    See references for more details.

    References:
    ----------
    .. [1] Pymatgen: Ong, S.P. et al. Comput. Mater. Sci. 68, 314-319 (2013).

    Examples:
    --------
    >>> import jaqpotpy as jt
    >>> comp = "Fe2O3"
    >>> featurizer = jt.descriptors.material.ElementPropertyFingerprint()
    >>> features = featurizer.featurize([comp])
    >>> type(features[0])
    <class 'numpy.ndarray'>
    >>> features[0].shape
    (65,)

    Note:
    ----
    This class requires matminer and Pymatgen to be installed.
    `NaN` feature values are automatically converted to 0 by this featurizer.

    """

    def __init__(self):
        """ """
        self.ep_featurizer: Any = None
        self.pstats = PropertyStats()

    @property
    def __name__(self):
        return "ElementPropertyFingerprint"

    def __getitem__(self):
        return self

    def _get_elemental_property(self, elem, property_name):
        if property_name == "block":
            block_key = {"s": 1.0, "p": 2.0, "d": 3.0, "f": 3.0}
            return block_key[getattr(elem, property_name)]
        else:
            value = getattr(elem, property_name)
            return np.nan if value is None else value

    def _featurize(self, datapoint: CompositionLike, **kwargs) -> np.ndarray:
        """Calculate chemical fingerprint from crystal composition.

        Parameters
        ----------
        datapoint: pymatgen.core.Composition object
          Composition object.

        Returns
        -------
        feats: np.ndarray
          Vector of properties and statistics derived from chemical
          stoichiometry. Some values may be NaN.

        """
        stats = ["minimum", "maximum", "range", "mean", "std_dev"]
        features = [
            "X",
            "row",
            "group",
            "block",
            "atomic_mass",
            "atomic_radius",
            "mendeleev_no",
            "electrical_resistivity",
            "velocity_of_sound",
            "thermal_conductivity",
            "melting_point",
            "bulk_modulus",
            "coefficient_of_linear_thermal_expansion",
        ]

        if not isinstance(datapoint, Composition):
            data = Composition(datapoint)
        else:
            data = datapoint

        all_attributes = []

        # Get the element names and fractions
        elements, fractions = zip(*data.element_composition.items())

        for attr in features:
            try:
                elem_data = [self._get_elemental_property(e, attr) for e in elements]
            except:
                raise ValueError("Cannot find property of Element.")

            for stat in stats:
                all_attributes.append(self.pstats.calc_stat(elem_data, stat, fractions))

        return np.array(all_attributes)

    def _featurize_dataframe(
        self, datapoint: CompositionLike, **kwargs
    ) -> pd.DataFrame:
        """Calculate chemical fingerprint from crystal composition.

        Parameters
        ----------
        datapoint: pymatgen.core.Composition object
          Composition object.

        Returns
        -------
        feats: np.ndarray
          Vector of properties and statistics derived from chemical
          stoichiometry. Some values may be NaN.

        """
        stats = ["minimum", "maximum", "range", "mean", "std_dev"]
        features = [
            "X",
            "row",
            "group",
            "block",
            "atomic_mass",
            "atomic_radius",
            "mendeleev_no",
            "electrical_resistivity",
            "velocity_of_sound",
            "thermal_conductivity",
            "melting_point",
            "bulk_modulus",
            "coefficient_of_linear_thermal_expansion",
        ]

        if not isinstance(datapoint, Composition):
            data = Composition(datapoint)
        else:
            data = datapoint

        all_attributes = []

        # Get the element names and fractions
        elements, fractions = zip(*data.element_composition.items())

        for attr in features:
            try:
                elem_data = [self._get_elemental_property(e, attr) for e in elements]
            except:
                raise ValueError("Cannot find property of Element.")

            for stat in stats:
                all_attributes.append(self.pstats.calc_stat(elem_data, stat, fractions))

        columns = [i + "_" + j for j in features for i in stats]
        return pd.DataFrame(
            index=[0], columns=columns, data=np.array(all_attributes).reshape(1, -1)
        )



import logging
import numpy as np
from jaqpotpy.utils.types import RDKitBond, RDKitMol, List, OneOrMany
from jaqpotpy.descriptors.base_classes import MolecularFeaturizer
from jaqpotpy.cfg import config
from typing import Optional

from rdkit import RDLogger
from rdkit import Chem

RDLogger.DisableLog("rdApp.*")

logger = logging.getLogger(__name__)


class GraphMatrix:
    """This is class used to store data for MolGAN neural networks.

    Parameters
    ----------
    node_features: np.ndarray
      Node feature matrix with shape [num_nodes, num_node_features]
    edge_features: np.ndarray,
      Edge feature matrix with shape [num_nodes, num_nodes]

    Returns
    -------
    graph: GraphMatrix
      A molecule graph with some features.

    """

    def __init__(self, adjacency_matrix: np.ndarray, node_features: np.ndarray):
        self.adjacency_matrix = adjacency_matrix
        self.node_features = node_features

    @property
    def __name__(self):
        return "GraphMatrix"

    def __getitem__(self):
        return self

    # @property
    # def adjacency_matrix(self):
    #   return self.adjacency_matrix
    #
    # @property
    # def node_features(self):
    #   return self.node_features


class MolGanFeaturizer(MolecularFeaturizer):
    """Featurizer for MolGAN de-novo molecular generation [1]_.
    The default representation is in form of GraphMatrix object.
    It is wrapper for two matrices containing atom and bond type information.
    The class also provides reverse capabilities.

    Examples
    --------
    >>> import jaqpotpy as jp
    >>> from rdkit import Chem
    >>> rdkit_mol, smiles_mol = Chem.MolFromSmiles('CCC'), 'C1=CC=CC=C1'
    >>> molecules = [rdkit_mol, smiles_mol]
    >>> featurizer = jp.descriptors.molecular.MolGanFeaturizer()
    >>> features = featurizer.featurize(molecules)
    >>> len(features) # 2 molecules
    2
    >>> type(features[0])
    <class 'deepchem.feat.molecule_featurizers.molgan_featurizer.GraphMatrix'>
    >>> molecules = featurizer.defeaturize(features) # defeaturization
    >>> type(molecules[0])
    <class 'rdkit.Chem.rdchem.Mol'>

    """

    @property
    def __name__(self):
        return "MolGan"

    def __getitem__(self):
        return self

    def __init__(
            self,
            max_atom_count: int = 9,
            kekulize: bool = True,
            sanitize: bool = False,
            bond_labels: List[RDKitBond] = None,
            atom_labels: List[int] = None,
    ):
        """Parameters
        ----------
        max_atom_count: int, default 9
          Maximum number of atoms used for creation of adjacency matrix.
          Molecules cannot have more atoms than this number
          Implicit hydrogens do not count.
        kekulize: bool, default True
          Should molecules be kekulized.
          Solves number of issues with defeaturization when used.
        bond_labels: List[RDKitBond]
          List of types of bond used for generation of adjacency matrix
        atom_labels: List[int]
          List of atomic numbers used for generation of node features

        References
        ----------
        .. [1] Nicola De Cao et al. "MolGAN: An implicit generative model for
           small molecular graphs" (2018), https://arxiv.org/abs/1805.11973

        """
        self.max_atom_count = max_atom_count

        self.BOND_DIM = 5
        self.kekulize = kekulize
        self.sanitize = sanitize

        try:
            from rdkit import Chem
        except ModuleNotFoundError:
            raise ImportError("This class requires RDKit to be installed.")

        self.bond_mapping = {
            "SINGLE": 0,
            0: Chem.BondType.SINGLE,
            "DOUBLE": 1,
            1: Chem.BondType.DOUBLE,
            "TRIPLE": 2,
            2: Chem.BondType.TRIPLE,
            "AROMATIC": 3,
            3: Chem.BondType.AROMATIC,
        }

        # bond labels
        if bond_labels is None:
            self.bond_labels = [
                Chem.rdchem.BondType.ZERO,
                Chem.rdchem.BondType.SINGLE,
                Chem.rdchem.BondType.DOUBLE,
                Chem.rdchem.BondType.TRIPLE,
                Chem.rdchem.BondType.AROMATIC,
            ]
        else:
            self.bond_labels = bond_labels

        self.SMILE_CHARSET = ["C", "N", "O", "F", "B", "I", "H", "S", "P", "Cl", "Br"]

        # atom labels
        if atom_labels is None:
            self.atom_labels = [
                0,
                6,
                7,
                8,
                9,
                5,
                53,
                1,
                16,
                15,
                17,
                35,
            ]  # C,N,O,F,B, I, H, S, P, Cl, Br
        else:
            self.atom_labels = atom_labels

        # create bond encoders and decoders
        self.bond_encoder = {li: i for i, li in enumerate(self.bond_labels)}
        self.bond_decoder = {i: li for i, li in enumerate(self.bond_labels)}
        # create atom encoders and decoders
        self.atom_encoder = {li: i for i, li in enumerate(self.atom_labels)}
        self.atom_decoder = {i: li for i, li in enumerate(self.atom_labels)}

        self.MAX_ATOMS = len(self.atom_labels)

    def _featurize(self, datapoint: RDKitMol, **kwargs) -> Optional[GraphMatrix]:
        """Calculate adjacency matrix and nodes features for RDKitMol.
        It strips any chirality and charges

        Parameters
        ----------
        datapoint: rdkit.Chem.rdchem.Mol
          RDKit mol object.

        Returns
        -------
        graph: GraphMatrix
          A molecule graph with some features.

        """
        try:
            from rdkit import Chem
        except ModuleNotFoundError:
            raise ImportError("This method requires RDKit to be installed.")
        if "mol" in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )

        if self.kekulize:
            Chem.Kekulize(datapoint)

        adjacency = np.zeros(
            (self.BOND_DIM, self.max_atom_count, self.max_atom_count), "float32"
        )
        features = np.zeros((self.max_atom_count, self.MAX_ATOMS), "float32")

        # loop over each atom in molecule
        for atom in datapoint.GetAtoms():
            i = atom.GetIdx()
            atom_type = self.atom_encoder[atom.GetAtomicNum()]
            features[i] = np.eye(self.MAX_ATOMS)[atom_type]
            # loop over one-hop neighbors
            for neighbor in atom.GetNeighbors():
                j = neighbor.GetIdx()
                bond = datapoint.GetBondBetweenAtoms(i, j)
                bond_type_idx = self.bond_mapping[bond.GetBondType().name]
                adjacency[bond_type_idx, [i, j], [j, i]] = 1

        # Where no bond, add 1 to last channel (indicating "non-bond")
        # Notice: channels-first
        adjacency[-1, np.sum(adjacency, axis=0) == 0] = 1

        # Where no atom, add 1 to last column (indicating "non-atom")
        features[np.where(np.sum(features, axis=1) == 0)[0], -1] = 1

        # degree = np.sum(
        #     adjacency[: datapoint.GetNumAtoms(), : datapoint.GetNumAtoms()], axis=-1
        # )

        graph = GraphMatrix(adjacency, features)
        # features[np.where(np.sum(X, axis=1) == 0)[0], -1] = 1
        return graph
        # return graph if (degree > 0).all() else None

        # A = np.zeros(
        #     shape=(self.max_atom_count, self.max_atom_count), dtype=np.float32)
        # bonds = datapoint.GetBonds()
        # begin, end = [b.GetBeginAtomIdx() for b in bonds], [
        #     b.GetEndAtomIdx() for b in bonds
        # ]
        # bond_type = [self.bond_encoder[b.GetBondType()] for b in bonds]
        # A[begin, end] = bond_type
        # A[end, begin] = bond_type
        # adjacency[bond_type, [begin, end], [end, begin]] = 1
        # adjacency[-1, np.sum(adjacency, axis=0) == 0] = 1
        # degree = np.sum(
        #     A[:datapoint.GetNumAtoms(), :datapoint.GetNumAtoms()], axis=-1)
        # X = np.array(
        #     [
        #         self.atom_encoder[atom.GetAtomicNum()]
        #         for atom in datapoint.GetAtoms()
        #     ] + [0] * (self.max_atom_count - datapoint.GetNumAtoms()),
        #     dtype=np.int32,
        # )
        #
        # x_d = np.eye(len(self.atom_labels))[X]
        # # features[np.where(np.sum(x_d, axis=1) == 0)[0], -1] = 1
        # # features[i] = np.eye(ATOM_DIM)[atom_type]
        # # graph = GraphMatrix(A, X)
        # graph = GraphMatrix(adjacency, x_d)
        # # features[np.where(np.sum(X, axis=1) == 0)[0], -1] = 1
        # # return graph
        # return graph if (degree > 0).all() else None

    def _defeaturize(
            self, graph_matrix: GraphMatrix, sanitize: bool = True, cleanup: bool = True
    ) -> RDKitMol:
        """Recreate RDKitMol from GraphMatrix object.
        Same featurizer need to be used for featurization and defeaturization.
        It only recreates bond and atom types, any kind of additional features
        like chirality or charge are not included.
        Therefore, any checks of type: original_smiles == defeaturized_smiles
        will fail on chiral or charged compounds.

        Parameters
        ----------
        graph_matrix: GraphMatrix
          GraphMatrix object.
        sanitize: bool, default True
          Should RDKit sanitization be included in the process.
        cleanup: bool, default True
          Splits salts and removes compounds with "*" atom types

        Returns
        -------
        mol: RDKitMol object
          RDKitMol object representing molecule.

        """
        # try:
        #     from rdkit import Chem
        #     import torch
        # except ModuleNotFoundError:
        #     raise ImportError("This method requires RDKit to be installed.")

        if not isinstance(graph_matrix, GraphMatrix):
            return None

        node_labels = graph_matrix.node_features
        edge_labels = graph_matrix.adjacency_matrix

        # edge_labels, node_labels = torch.max(torch.from_numpy(edge_labels), -1)[1].cpu().numpy()\
        #   , torch.max(torch.from_numpy(node_labels), -1)[1].cpu().numpy()

        mol = Chem.RWMol()

        # Remove "no atoms" & atoms with no bonds
        keep_idx = np.where(
            (np.argmax(node_labels, axis=1) != len(self.atom_labels) - 1)
            & (np.sum(edge_labels[:-1], axis=(0, 1)) != 0)
        )[0]
        features = node_labels[keep_idx]
        adjacency = edge_labels[:, keep_idx, :][:, :, keep_idx]

        # Add atoms to molecule
        for atom_type_idx in np.argmax(features, axis=1):
            atomic_number = self.atom_decoder[atom_type_idx]
            atom = Chem.Atom(atomic_number)
            mol.AddAtom(atom)

        # Add bonds between atoms in molecule; based on the upper triangles
        # of the [symmetric] adjacency tensor
        # (bonds_ij, atoms_i, atoms_j) = np.where(np.triu(adjacency) == 1)
        # for (bond_ij, atom_i, atom_j) in zip(bonds_ij, atoms_i, atoms_j):
        #     if atom_i == atom_j or bond_ij == len(self.bond_labels) - 1:
        #         continue
        #     bond_type = self.bond_decoder[bond_ij]
        #     try:
        #       mol.AddBond(int(atom_i), int(atom_j), bond_type)
        #     except Exception as e:
        #       continue

        (bonds_ij, atoms_i, atoms_j) = np.where(np.triu(adjacency) == 1)
        for bond_ij, atom_i, atom_j in zip(bonds_ij, atoms_i, atoms_j):
            if atom_i == atom_j or bond_ij == self.BOND_DIM - 1:
                continue
            bond_type = self.bond_mapping[bond_ij]
            mol.AddBond(int(atom_i), int(atom_j), bond_type)

        # mol = Chem.RWMol()
        #
        # for node_label in node_labels:
        #   mol.AddAtom(Chem.Atom(self.atom_decoder[node_label]))
        #
        # for start, end in zip(*np.nonzero(edge_labels)):
        #   if start > end:
        #     mol.AddBond(int(start), int(end), self.bond_decoder[edge_labels[start, end]])

        if sanitize:
            try:
                Chem.SanitizeMol(mol)
            except Exception:
                mol = None

        if cleanup:
            try:
                smiles = Chem.MolToSmiles(mol)
                smiles = max(smiles.split("."), key=len)
                if "*" not in smiles:
                    mol = Chem.MolFromSmiles(smiles)
                else:
                    mol = None
            except Exception:
                mol = None

        return mol

    def defeaturize(
            self, graphs: OneOrMany[GraphMatrix], log_every_n: int = 1000
    ) -> np.ndarray:
        """Calculates molecules from corresponding GraphMatrix objects.

        Parameters
        ----------
        graphs: GraphMatrix / iterable
          GraphMatrix object or corresponding iterable
        log_every_n: int, default 1000
          Logging messages reported every `log_every_n` samples.

        Returns
        -------
        features: np.ndarray
          A numpy array containing RDKitMol objext.

        """
        # Special case handling of single molecule
        if isinstance(graphs, GraphMatrix):
            graphs = [graphs]
        else:
            # Convert iterables to list
            graphs = list(graphs)

        molecules = []
        for i, gr in enumerate(graphs):
            if i % log_every_n == 0:
                logger.info("Featurizing datapoint %i" % i)
            try:
                molecules.append(self._defeaturize(gr))
            except Exception as e:
                if config.verbose is True:
                    logger.warning(
                        "Failed to defeaturize datapoint %d, %s. Appending empty array",
                        i,
                        gr,
                    )
                    logger.warning("Exception message: {}".format(e))
                else:
                    continue
                molecules.append(np.array([]))

        return np.asarray(molecules)

    def _get_column_names(self):
        return ["MolGanGraphs"]

    def _featurize_dataframe(self, datapoint: RDKitMol, **kwargs):
        return self._featurize(datapoint, **kwargs)



class PretrainedRequirements:
    def __init__(
        self,
        rawModel=None,
        independentFeatures=None,
        predictedFeatures=None,
        additionalInfo=None,
        title=None,
        implementedWith=None,
        algorithm=None,
        implementedIn=None,
        description=None,
        dependentFeatures=None,
        jaqpotpyVersion=None,
        jaqpotpyDockerVersion=None,
        type=None,
        libraries=None,
        versions=None,
    ):
        self.rawModel = rawModel
        self.independentFeatures = independentFeatures
        self.predictedFeatures = predictedFeatures
        self.dependentFeatures = dependentFeatures
        self.additionalInfo = additionalInfo
        self.title = title
        self.implementedWith = implementedWith
        self.implementedIn = implementedIn
        self.algorithm = algorithm
        self.description = description
        self.jaqpotpyVersion = jaqpotpyVersion
        self.jaqpotpyDockerVersion = jaqpotpyDockerVersion
        self.type = type
        self.libraries = libraries
        self.versions = versions
